/* ====================================================================
A MODEL OF COMPOSITION OF NETWORKS                        10 March 2023 
==================================================================== */
open net[]
open traffic[]
open group []

/* --------------------------------------------------------------------
COMPOSITION OF NETWORKS BY BRIDGING
   A Bridging object represents the semantics of a set of networks that
are able to communicate across bridging links.  
   Because Bridging objects are used where bridging is optional, the
semantics allows Bridging objects with one network and no bridging
links, i.e., vestigial cases of bridging.
   Bridging requires the convention that links are globally unique.
Unlike the convention for member names, this convention is no trouble,
because links are not passed around in packet headers.
-------------------------------------------------------------------- */

sig Bridging {
   nets: set NetworkStateWithGroups,
   reachableWithBridging: NetHdr -> Name -> Name             -- derived
}  {
-- Consistency of name spaces among bridged networks.  These rules are
-- elaborate because they must insure that group names, as well as
-- member names, are used consistently.
   -- Member names cannot overlap.
      all disj n, n": nets | no (n.members & n".members) 
   -- If a group name is used by multiple networks, then either the
   -- group is shared (members of the group are the the same in all the
   -- networks) or the group is local (each group's members are 
   -- confined to one network).
      all g: Name, disj n, n": nets | (g in n.groups && g in n".groups)
      => (  -- nonlocal group
            (  (  (g in n.allGroups && g in n".allGroups)
                  || (g in n.anyGroups && g in n".anyGroups)  )
            && g.(n.groupSenders) = g.(n".groupSenders)
            && g.(n.groupReceivers) = g.(n".groupReceivers)   )
         || -- local group
            (  g.(n.(groupSenders + groupReceivers)) in n.members
            && g.(n".(groupSenders + groupReceivers)) in n".members  )
         )
   -- If a group member is a member of a network, then the network must
   -- declare the group.
      all n: nets, m: n.members, g: Name |
         m in g.(nets.(groupSenders + groupReceivers)) => g in n.groups
   -- A name cannot be used as a group name in one network and a member
   -- name in another network.
      all disjoint n, n": nets | (no n.groups & n".members)
-- Links are globally unique, so if the link sets of two networks
-- overlap, then the links are bridging links.
   all disj n, n": nets, k: nets.links | 
      (k in n.links && k in n".links)
      <=> (  some ( n.members & (k.sndr + k.rcvrs) )
          && some ( n".members & (k.sndr + k.rcvrs) )  )
-- Reachability.
   reachableWithBridging =
   {  h: NetHdr, m, m": nets.members |
         some k, k": nets.links, kin, kout: LinkIdent |
            (m -> kout -> k) in nets.outLinks
         && (m" -> kin -> k") in nets.inLinks
         && (h -> kout) in
               (LinkIdent + Self).((m.(nets.forwardTables)).frows)
         && (kin -> h) in 
               ((m".(nets.acquireTables)).arows).(Receive+Forward)
         && (  k = k" || (k -> k") in ^(h.(nets.oneStep))  )         }
} 

/* --------------------------------------------------------------------
BRIDGING WITH A TRAFFIC MODEL
-------------------------------------------------------------------- */

sig BridgingWithTraffic in Bridging {
   effectivelyReachableWithBridging: NetHdr -> Name -> Name  -- derived
}  {
   nets in NetworkStateWithTraffic
-- Derivation of effective reachability.
   effectivelyReachableWithBridging =
   {  h: NetHdr, m, m": nets.members |
         some k, k": nets.links |
            (m -> h -> k) in nets.effectiveSend
         && (m" -> h -> k") in nets.effectiveReceive
         && (  k = k" || (k -> k") in ^(h.(nets.oneStep))  )  }
}

/* --------------------------------------------------------------------
COMPOSITION OF NETWORKS BY LAYERING
   Layering does not change the semantics of either the overlay or any
of its underlays.  Layering implements links in networks with sessions
in other networks, so that the links do not have to be implemented
physically.
-------------------------------------------------------------------- */

sig Layering {
   overs: Bridging,
   unders: BridgingWithTraffic,
   implementations: Link -> lone SessionIdent,
   attachments: Name -> NetworkState -> Name,
   groupAttachments: Name lone -> lone Name,
   oneImplemented: set Link,                                 -- derived
   twoImplemented: set Link,                                 -- derived
   groupImplemented: set Link                                -- derived
}  {
-- Type and configuration constraints.
   -- A network cannot be layered on itself.
      no (overs.nets & unders.nets)
   -- Implemented links belong to the overlays.
      implementations.SessionIdent in (overs.nets).links
   -- Implemented links have attached endpoints.  The only attached
   -- members are the endpoints of implemented links.
      (implementations.SessionIdent).(sndr + rcvrs)
         = (attachments.Name).NetworkState
   -- A name is attached to a unique underlay network and member of it.
      all m: (attachments.Name).NetworkState | 
      let attachMem = NetworkState.(m.attachments) |
      let attachNet = (m.attachments).Name |
         (  one attachNet && one attachMem
         && attachNet in unders.nets 
         && attachMem in attachNet.members  )
   -- An underlay network member is the attachment of a unique overlay
   -- name.
      all n: NetworkState, m: Name | lone (attachments.m).n
-- Derivation of link sets.  These constraints complete the one-to-one
-- correspondence between implemented links and implementing sessions.
   oneImplemented = ((implementations.SessionIdent) <: mode).OneWay
   -- OneLinks come as singles.
      all s: oneImplemented.implementations | one implementations.s 
   twoImplemented = ((implementations.SessionIdent) <: mode).TwoWay
   -- TwoLinks come in pairs.
      all s: twoImplemented.implementations | 
         (  # implementations.s = 2
         && some disj k, k': implementations.s | ReverseLink [k, k']  )
   groupImplemented = 
      ((implementations.SessionIdent)<:mode).(AllCast + AnyCast)
   -- AnyCast links come as singles.
      all s: (groupImplemented & mode.AnyCast).implementations | 
         one implementations.s 
   -- Multiple allCast links can be implemented by the same session, 
   -- but only if they have the same group.
      all disj k, k": (groupImplemented & mode.AllCast) |
         k.implementations = k".implementations => k.group = k".group
-- Additional type and configuration constraints for groups.
   -- If some of a group's links are implemented, the group must be 
   -- nonlocal in this bridging.
      all g: groupAttachments.Name, disj n, n": overs.nets | 
         (g in n.groups && g in n".groups)
      => (  g.(n.groupSenders) = g.(n".groupSenders)
         && g.(n.groupReceivers) = g.(n".groupReceivers)  )
   -- The members of an implemented group map into group membership in 
   -- the underlay.
      all k: groupImplemented, g: k.group |
      (  g in (overs.nets).groups
      && g in groupAttachments.Name
      && (  g in (overs.nets).allGroups 
         <=> g.groupAttachments in (unders.nets).allGroups  )  
      && (  g in (overs.nets).anyGroups 
         <=> g.groupAttachments in (unders.nets).anyGroups  )  
      && NetworkState.((k.sndr).attachments) in
            (g.groupAttachments).((unders.nets).groupSenders)
      && NetworkState.((k.rcvrs).attachments) in
            (g.groupAttachments).((unders.nets).groupReceivers) 
      )
-- Valid overlays.
   -- The overlay transmit tables invoke the implementations.
      all k: implementations.SessionIdent,    -- link to be implemented
          s: k.implementations,                -- identifier of session
          n: ((k.sndr).attachments).Name |           -- sender attached
                                                     -- in network n
          let x =(k.sndrIdent).((k.sndr).((overs.nets).transmitTable))|
             x in ExternalSession        -- x is entry in transmitTable
          && x.exNet = n.networkName && x.exSession = s 
-- Valid underlays.
   -- The formation of an implementing session, in one direction.  The 
   -- session s has a header h for the sender in this direction.  Src
   -- and dst names in the header match the link-endpoint attachments.
   -- In its underlay network, the sender sends.  In its underlay
   -- network, which may be the same or different, the receiver 
   -- receives.
   all k:implementations.SessionIdent, s:k.implementations, r:k.rcvrs |
      let underSndrNet = ((k.sndr).attachments).Name                |
      let underSndrMem = NetworkState.((k.sndr).attachments)        |
      let underRcvrNet = (r.attachments).Name                       |
      let underRcvrMem = NetworkState.(r.attachments)               |
      let underRcvrGrp = (k.group).groupAttachments                 |
      let overRcvrNetName = ((overs.nets <: members).r).networkName | 
      some h: NetHdr | 
         (  h.src = underSndrMem
         && (k in groupImplemented 
               =>   h.dst = underRcvrGrp
               else h.dst = underRcvrMem)
         && h.sessionIdent = s 
         && h.overlay = overRcvrNetName 
         && (  k in oneImplemented   => h.protocol in TwoEndOneWay
               else (k in twoImplemented => h.protocol in TwoEndTwoWay 
               else h.protocol in MultiWay                        )  )
         && (h.src -> h) in (underSndrNet.effectiveSend).Link
         && (underRcvrMem -> h) in (underRcvrNet.effectiveReceive).Link
         && (some x: ExternalLink |
                 x.exNet = overRcvrNetName 
              && x.exLink = r.(k.rcvrIdents)
              && (underRcvrMem -> h -> x) in underRcvrNet.receiveTable)
         )                         
}

/* --------------------------------------------------------------------
PROPERTIES OF LAYERING, GLOBAL CONSTRAINTS ON LAYERING
   Layering objects are operational, in the sense that table entries 
are derived from them.  The constraints here are expressed as facts 
rather than in Layering signatures because they constrain multiple 
layering objects.
-------------------------------------------------------------------- */

pred Underlay_satisfies_layering_demands [y: Layering] {
-- The communication demands of an overlay can be identified by the
-- session identifiers in the layering.  This property is the desired
-- semantics of layering.
   all k: (y.implementations).SessionIdent, s: k.(y.implementations), 
                                                    h: sessionIdent.s |
      k ! in y.groupImplemented 
      => (h -> h.src -> h.dst) in 
            (y.unders).effectivelyReachableWithBridging
      else all r: k.rcvrs | some h: sessionIdent.s |
            (  h.dst = k.group 
            && (h -> h.src -> NetworkState.(r.(y.attachments)))
                  in (y.unders).effectivelyReachableWithBridging  )   }

fact Layering_is_hierarchical {
-- The overlay/underlay dependence relation must be a directed acyclic 
-- graph.
   let hierarchy = { n, n": NetworkState | some y: Layering | 
      n in y.overs.nets && n" in y.unders.nets              } |
         all n: NetworkState | n ! in n.(^hierarchy)          }

fact Implementation_is_unique {
-- No link can be implemented in more than one layering operation.
   all disj y, y": Layering | 
      some (y.overs.nets & y".overs.nets)
   => no (y.implementations).SessionIdent & 
         (y".implementations).SessionIdent  }

fact Implementation_user_is_unique {
-- No session can implement more than one overlay link.
   all disj y, y": Layering | 
      some (y.unders.nets & y".unders.nets)
   => no Link.(y.implementations) & Link.(y".implementations)  }

fact Underlay_member_attaches_unique_overlay_member {
-- An underlay member cannot be the attachment of more than one member
-- of the same overlay network.
   no disj y, y": Layering |
      some (y.overs.nets & y".overs.nets)
   && (  some disj m, m": Name, n: NetworkState, a: Name |
            (m -> n -> a) in y.attachments 
         && (m" -> n -> a) in y".attachments             )  }

fact Overlay_member_attaches_to_unique_underlay_member {
-- An overlay member cannot be attached to multiple members of the 
-- same underlay network.
   no disj y, y": Layering |
      some (y.overs.nets & y".overs.nets)
   && (  some m: Name, n: NetworkState, disj a, a": Name |
            (m -> n -> a) in y.attachments 
         && (m -> n -> a") in y".attachments             )  }

/* --------------------------------------------------------------------
COMPOSITION OF NETWORKS BY SUBDUCTION
   Subduction is a special kind of composition needed when a network
(or set of bridged networks) is both bridged with and layered on the 
same network or networks.  It has the fields and constraints of a 
Layering operator, augmented with the extended reachability relations 
of a Bridging operator.
   Shared members are the single exception to the global uniqueness of 
member names.  Shared members have a single name, under which they are
members of both an overlay and an underlay network.  This is necessary
because they share the role of being endpoints of the same shared link.
   The tables at both levels operationalize the sharing of links, as 
follows:
   Shared incoming link to overlay network:
      Underlay acquire can yield SharedAcquire.  If so, the same inLink
      and header cannot also yield Receive or Forward, and the packet
      is passed immediately to the acquireTable in the overlay.
   Shared outgoing link from overlay network:
      The overlay can forward normally onto a shared outLink.  However,
      often the overlay does not have the knowledge to choose a shared
      outLink, and wants to delegate the choice to the underlays.  To
      do this, the overlay forwardTable, instead of an outLink, has
      SharedForward.  With this entry, the packet is sent immediately 
      to the forwardTable in the underlay, with its incoming link set
      to the value SharedPacket.  From this incoming link, the underlay
      can only forward onto a shared outLink.
Note that any Layering object can be extended to a Subduction object.
If there are no shared members, then the Subduction is vestigial.
-------------------------------------------------------------------- */

sig Subduction extends Layering {
   sharedLinks: set Link,                                    -- derived
   sharedMembers: set Name,                                  -- derived
   subductionOverlaysOneStep: NetHdr -> Link -> Link,        -- derived
   reachableWithSubduction: NetHdr -> Name -> Name           -- derived
}  {
-- Derivation of and constraints on shared links.
   -- Shared links are the links shared with the underlays.  Note that 
   -- a shared link might also be a bridging link between two underlay 
   -- networks, or an external link of the underlay network set.
      sharedLinks = (unders.nets).links & (overs.nets).links
   -- Shared links must be external links of the bridged overlay 
   -- networks, viewed as a single network.
      sharedLinks in
         { k: (overs.nets).links | 
            (k.sndr + k.rcvrs) ! in (overs.nets).members }
   -- A shared link must not be a group link.
      no sharedLinks.group
   -- A shared inLink has the same linkIdent at both levels. 
      all k: sharedLinks | 
         k.rcvrs in sharedMembers =>
            (k.rcvrs).((overs.nets).inLinks).k 
          = (k.rcvrs).((unders.nets).inLinks).k
   -- A shared outLink has the same linkIdent at both levels. 
      all k: sharedLinks | 
         k.sndr in sharedMembers =>
            (k.sndr).((overs.nets).outLinks).k
          = (k.sndr).((unders.nets).outLinks).k
-- Derivation of and constraints on shared members.
   -- Shared members are the endpoints in the overlays of shared links.
      sharedMembers = 
         (overs.nets).members & (sharedLinks).(sndr + rcvrs)
   -- Shared members of the overlays are attached to members with the
   -- same name in the underlays.  This is necessary because shared 
   -- links have a single endpoint in two different networks, and a
   -- single endpoint of a link can only have one name.
      all m: sharedMembers | NetworkState.(m.attachments) = m
   -- Except for shared members, member names in the overlays do not
   -- overlap with member names in the underlays.
      no (((overs.nets).members-sharedMembers) & (unders.nets).members)
   -- If the overlay networks have an external link whose external 
   -- endpoint is a member of the underlays, then it must be a shared 
   -- link.
      all k: (overs.nets).links, m: k.(sndr + rcvrs) |
         (m ! in (overs.nets).members && m in (unders.nets).members) 
      => k in sharedLinks
-- Constraints on layering.
   -- Any unshared link of the overlay networks can be implemented by 
   -- layering.
      implementations.SessionIdent in ((overs.nets).links -sharedLinks)
   -- There must be some implemented overlay links, because if not,
   -- this should simply be bridging, not subduction.
      some implementations
-- Constraints on network tables.
   -- If an underlay acquireTable has an entry SharedAcquire, then the
   -- inLink must be a shared link.  Because of this table entry, the
   -- incoming packets matching the table entry will be handled by the
   -- overlay.  The packet cannot also be received or forwarded in the
   -- underlay.
      all m: Name, h: NetHdr, i: LinkIdent |
      SharedAcquire in h.(i.((m.((unders.nets).acquireTables)).arows))
      => (  m in sharedMembers
         && i.(m.((unders.nets).inLinks)) in sharedLinks 
         && one h.(i.((m.((unders.nets).acquireTables)).arows))  )
   -- In overlay forwarding, if a forwardTable says SharedForward, then
   -- the table must be located at a shared member with an outgoing
   -- shared link.
      all m: (overs.nets).members |
         SharedForward in 
         NetHdr.((LinkIdent + NoInLink).
            ((m.((overs.nets).forwardTables)).frows))
         => (  m in sharedMembers
            && (some LinkIdent.(m.((overs.nets).outLinks)) 
                     & sharedLinks)                      )
   -- In underlay forwarding, packets from inLink SharedPacket must be
   -- forwarded onto a shared outLink.
      all m: Name, h: NetHdr, i: LinkIdent |
         (SharedPacket -> h -> i) in 
            (m.((unders.nets).forwardTables)).frows
         => (  m in sharedMembers
            && i.(m.((unders.nets).outLinks)) in (sharedLinks + Drop) )
-- Reachability.
   -- Add extra tuples for delegated forwarding.  Remove tuples if
   -- acquire from a shared link is not suppored by underlay acquire.
      subductionOverlaysOneStep = (overs.nets).oneStep
      +  {  h: NetHdr, disj k, k": (overs.nets).links |
            some m: sharedMembers, kin, kout: LinkIdent |
               (m -> kin -> k) in (overs.nets).inLinks
            && (m -> kout -> k") in (overs.nets).outLinks
            && (kin -> h -> Forward) in 
                  (m.((overs.nets).acquireTables)).arows
            && (kin -> h -> SharedForward) in 
                  (m.((overs.nets).forwardTables)).frows  
            && (SharedPacket -> h -> kout) in 
                  (m.((unders.nets).forwardTables)).frows            }
      -  {  h: NetHdr, disj k, k": (overs.nets).links |
            some m: sharedMembers, kin: LinkIdent |
               (m -> kin -> k) in (overs.nets).inLinks
            && k in sharedLinks
            && (kin -> h -> SharedAcquire) ! in 
                  (m.((unders.nets).acquireTables)).arows  }
      reachableWithSubduction = 
         {  h: NetHdr, m, m": Name |
            let allNets = overs.nets + unders.nets |
            some k, k": allNets.links, kin, kout: LinkIdent |
               (some n: allNets |
                  (m -> kout -> k) in n.outLinks
               && (  (h -> kout) in
                        (LinkIdent+Self).((m.(n.forwardTables)).frows)
                  || (  n in overs.nets
                     && (h -> SharedForward) in 
                        (LinkIdent+Self).((m.(n.forwardTables)).frows)
                     && (SharedPacket -> h -> kout) in 
                           (m.((unders.nets).forwardTables)).frows 
                     )
                  )
               )
            && (some n: allNets |
                  (m" -> kin -> k") in n.inLinks
               && (  (kin -> h) in
                       ((m".(n.acquireTables)).arows).(Receive+Forward)
                  && (  (  n in overs.nets && m" in sharedMembers
                        && k" in sharedLinks                    ) 
                    => (kin -> h -> SharedAcquire) in 
                           (m".((unders.nets).acquireTables)).arows 
                     )
                  )
               )
            && (  k = k" 
               || (k -> k") in 
                 ^(h.(subductionOverlaysOneStep+(unders.nets).oneStep))
               )  
         }
}

fact Shared_members_are_unique {
-- No network member can be a shared member in more than one Subduction
-- operator.
   all disj s, s": Subduction, m: s.sharedMembers, n: NetworkState |
   (  n in (s.overs+s.unders).nets && m in n.members  )
   => (  n !in (s".overs+s".unders).nets || m !in s".sharedMembers  ) }

/* --------------------------------------------------------------------
SUBDUCTION WITH TRAFFIC MODELS
-------------------------------------------------------------------- */

sig SubductionWithTraffic in Subduction {
   effectivelyReachableWithSubduction: NetHdr -> Name -> Name --derived
}  {
   overs in BridgingWithTraffic
-- Derivation of effective reachability.
   effectivelyReachableWithSubduction =
      {  h: NetHdr, m, m": Name |
         let allNets = overs.nets + unders.nets |
         some k, k": allNets.links, kin, kout: LinkIdent |
            (some n: allNets |
               (m -> kout -> k) in n.outLinks
            && (m -> h -> k) in n.effectiveSend
            && (  (h -> kout) in
                     (LinkIdent+Self).((m.(n.forwardTables)).frows)
               || (  n in overs.nets
                  && (h -> SharedForward) in 
                     (LinkIdent+Self).((m.(n.forwardTables)).frows)
                  && (SharedPacket -> h -> kout) in 
                        (m.((unders.nets).forwardTables)).frows 
                  )
               )
            )
         && (some n: allNets |
               (m" -> kin -> k") in n.inLinks
            && (m" -> h -> k") in n.effectiveReceive
            && (  (kin -> h) in
                    ((m".(n.acquireTables)).arows).(Receive+Forward)
               && (  (  n in overs.nets && m" in sharedMembers
                     && k" in sharedLinks                    ) 
                 => (kin -> h -> SharedAcquire) in 
                        (m".((unders.nets).acquireTables)).arows 
                  )
               )
            )
         && (  k = k" 
            || (k -> k") in 
                 ^(h.(subductionOverlaysOneStep+(unders.nets).oneStep))
            )  
      }
}

/* ====================================================================
run P for 1 but
   N>=S NetworkName, S NetworkState,
   B Bridging, Y Layering, U Subduction,
   M Name, L<=K LinkIdent, K Link, D SessionIdent, D-2D NetHdr,
   P Protocol,
   SxM AcquireTable, SxM ForwardTable, X ExternalReference
==================================================================== */
/* ====================================================================
VERIFICATION ARCHIVE
==================================================================== */
/* --------------------------------------------------------------------
TESTS OF BRIDGING
-------------------------------------------------------------------- */

pred BridgingTest1 [b: Bridging] {
-- Within two bridged networks, all members can be reachable from all
-- other members.
   some disj n, n": b.nets | {
      # n.members = 2 && # n".members = 2
      all disj m, m": n.members + n".members |
         some h: NetHdr | (h -> m -> m") in b.reachableWithBridging  }}
run BridgingTest1 for 1 but
   2 NetworkName, 2 NetworkState, 
   4 Name, 2 LinkIdent, 4 Link, 6 SessionIdent, 12 NetHdr,
   4 AcquireTable, 4 ForwardTable

pred BridgingTest2 [b: Bridging] {  one b.nets  }
-- The bridging set can be a singleton.
run BridgingTest2 for 1 but
   2 NetworkName, 2 NetworkState, 
   4 Name, 3 LinkIdent, 4 Link, 1 SessionIdent, 1 NetHdr,
   4 AcquireTable, 4 ForwardTable

pred BridgingTest3 [b: Bridging] {  no b.nets  }
-- The bridging set can be empty.
run BridgingTest3 for 1 but
   2 NetworkName, 2 NetworkState, 
   4 Name, 3 LinkIdent, 4 Link, 1 SessionIdent, 1 NetHdr,
   4 AcquireTable, 4 ForwardTable

pred Useful_bridging_over_reachable_test [b: Bridging] { 
-- Bridging can add to reachability.
      # b.nets >= 2
      some (b.nets).reachable
      some (b.reachableWithBridging - (b.nets).reachable)  }
run Useful_bridging_over_reachable_test for 1 but
   2 NetworkName, 2 NetworkState, 
   4 Name, 3 LinkIdent, 4 Link, 1 SessionIdent, 1 NetHdr,
   4 AcquireTable, 4 ForwardTable

pred All_bridging_links_are_allCast [b: Bridging] {
   some disj n, n": b.nets |
   (  (n + n") in NetworkStateWithGroups  
   && all k: n.links & n".links | k.mode = AllCast  )  }

pred BridgingTest1WithGroups [b: Bridging] {
   BridgingTest1 [b]
   All_bridging_links_are_allCast [b] }
run BridgingTest1WithGroups for 1 but
   2 NetworkName, 2 NetworkState, 
   5 Name, 2 LinkIdent, 4 Link, 6 SessionIdent, 4 NetHdr,
   4 AcquireTable, 4 ForwardTable

pred Useful_group_bridging_over_reachable_test [b: Bridging] {
   Useful_bridging_over_reachable_test [b]
   All_bridging_links_are_allCast [b]                        }
run Useful_group_bridging_over_reachable_test for 1 but
   2 NetworkName, 2 NetworkState, 
   5 Name, 3 LinkIdent, 4 Link, 4 SessionIdent, 4 NetHdr,
   4 AcquireTable, 4 ForwardTable

pred Bridging_with_local_groups [b: Bridging] {
   some (b.nets).reachable
   some (b.reachableWithBridging - (b.nets).reachable) 
   some disj n, n": b.nets, g: n.allGroups & n".anyGroups |
      some g.(n.groupReceivers) && some g.(n".groupReceivers)  } 
run Bridging_with_local_groups for 1 but
   2 NetworkName, 2 NetworkState, 
   6 Name, 4 LinkIdent, 6 Link, 4 SessionIdent, 4 NetHdr,
   4 AcquireTable, 4 ForwardTable

pred Bridging_with_a_shared_group [b: Bridging] {
   BridgingTest1 [b]
   All_bridging_links_are_allCast [b]
   one (b.nets).allGroups 
   all disj n, n": b.nets | n.allGroups = n".allGroups  }
run Bridging_with_a_shared_group for 1 but
   2 NetworkName, 2 NetworkState, 
   5 Name, 4 LinkIdent, 6 Link, 4 SessionIdent, 4 NetHdr,
   4 AcquireTable, 4 ForwardTable

/* --------------------------------------------------------------------
THEOREMS ABOUT BRIDGING
-------------------------------------------------------------------- */

assert Vestigial_bridging_over_reachable_theorem {
-- If there are no bridging links, bridging adds no reach.
   all b: Bridging | 
      (all disj n, n": b.nets | no (n.links & n".links))
   => b.reachableWithBridging = (b.nets).reachable }
check Vestigial_bridging_over_reachable_theorem for 1 but
   2 NetworkName, 2 NetworkState, 
   4 Name, 3 LinkIdent, 4 Link, 2 SessionIdent, 2 NetHdr,
   4 AcquireTable, 4 ForwardTable

assert reachable_in_bridging_inclusion_theorem {
-- Bridging never diminishes the reach of a network.
   all b: Bridging | 
      (b.nets).reachable in b.reachableWithBridging  }
check reachable_in_bridging_inclusion_theorem for 1 but
   2 NetworkName, 2 NetworkState,
   4 Name, 3 LinkIdent, 6 Link, 2 SessionIdent, 2 NetHdr,
   4 AcquireTable, 4 ForwardTable

assert Bridging_closure_theorem {
-- This theorem shows that, for any two networks, there is a third
-- network that is equivalent to the first two networks bridged
-- together. 
   all b: Bridging, n0, n1: b.nets, n2: NetworkState | {
      b.nets = n0 + n1
      n2.infras = n0.infras + n1.infras
      n2.users = n0.users + n1.users
      n2.oneLinks = n0.oneLinks + n1.oneLinks
      n2.twoLinks = n0.twoLinks + n1.twoLinks
      n2.acquireTables = n0.acquireTables + n1.acquireTables
      n2.forwardTables = n0.forwardTables + n1.forwardTables
      all n: n0 + n1 + n2 | n.links = n.oneLinks + n.twoLinks
   }  => 
      n2.reachable = b.reachableWithBridging                }
check Bridging_closure_theorem for 1 but
   3 NetworkName, 3 NetworkState,
   2 Name, 1 LinkIdent, 2 Link, 2 SessionIdent, 2 NetHdr,
   6 AcquireTable, 6 ForwardTable
check Bridging_closure_theorem for 1 but
   3 NetworkName, 3 NetworkState,
   4 Name, 2 LinkIdent, 4 Link, 3 SessionIdent, 3 NetHdr,
   8 AcquireTable, 8 ForwardTable

/* --------------------------------------------------------------------
TESTS AND THEOREMS ABOUT BRIDGING WITH A TRAFFIC MODEL
-------------------------------------------------------------------- */

pred EffectiveBridgingTest [b: BridgingWithTraffic] {
-- Within two bridged networks, all members can be effectively 
-- reachable from all other members.
   some disj n, n": b.nets | {
      # n.members = 2 && # n".members = 2
      all disj m, m": n.members + n".members |
         some h: NetHdr | 
            (h -> m -> m") in b.effectivelyReachableWithBridging  }  }
run EffectiveBridgingTest for 1 but
   2 NetworkName, 2 NetworkState, 
   4 Name, 2 LinkIdent, 4 Link, 6 SessionIdent, 12 NetHdr,
   4 AcquireTable, 4 ForwardTable

pred Useful_ERWB_over_ER_test [b: BridgingWithTraffic] { 
-- Effective bridging can add to effective reachability.
      # b.nets >= 2
      some (b.nets).effectivelyReachable
      some (b.effectivelyReachableWithBridging 
            - (b.nets).effectivelyReachable  )               }
run Useful_ERWB_over_ER_test for 1 but
   2 NetworkName, 2 NetworkState, 
   4 Name, 3 LinkIdent, 4 Link, 2 SessionIdent, 2 NetHdr,
   4 AcquireTable, 4 ForwardTable

assert Vestigial_ERWB_over_ER_theorem {
-- If there are no bridging links, effective bridging adds no reach.
   all b: BridgingWithTraffic | 
      (all disj n, n": b.nets | no (n.links & n".links))
   => b.effectivelyReachableWithBridging 
       = (b.nets).effectivelyReachable                 }
check Vestigial_ERWB_over_ER_theorem for 1 but
   2 NetworkName, 2 NetworkState, 
   4 Name, 3 LinkIdent, 4 Link, 2 SessionIdent, 2 NetHdr,
   4 AcquireTable, 4 ForwardTable

assert ER_in_ERWB_inclusion_theorem {
   all b: BridgingWithTraffic |
      (b.nets).effectivelyReachable 
         in b.effectivelyReachableWithBridging  }
check ER_in_ERWB_inclusion_theorem for 1 but
   2 NetworkName, 2 NetworkState, 
   5 Name, 3 LinkIdent, 5 Link, 3 SessionIdent, 3 NetHdr,
   5 AcquireTable, 5 ForwardTable

pred Useful_RWB_over_ERWB_test [b: BridgingWithTraffic] {
-- This shows that reachableWithBridging is bigger than its effective
-- counterpart.
   some b.effectivelyReachableWithBridging
   some (b.reachableWithBridging - b.effectivelyReachableWithBridging)}
run Useful_RWB_over_ERWB_test for 1 but
   2 NetworkName, 2 NetworkState, 
   5 Name, 3 LinkIdent, 5 Link, 2 SessionIdent, 2 NetHdr,
   5 AcquireTable, 5 ForwardTable

assert ERWB_in_RWB_inclusion_theorem {
   all b: BridgingWithTraffic |
      b.effectivelyReachableWithBridging in b.reachableWithBridging  }
check ERWB_in_RWB_inclusion_theorem for 1 but
   2 NetworkName, 2 NetworkState, 
   5 Name, 3 LinkIdent, 5 Link, 1 SessionIdent, 1 NetHdr,
   5 AcquireTable, 5 ForwardTable

pred reachable_and_ERWB_unordered [b: BridgingWithTraffic] {
-- This shows that reachable and effectivelyReachableWithBridging are 
-- not ordered with respect to each other.
      some (b.effectivelyReachableWithBridging - (b.nets).reachable)
   && some ((b.nets).reachable - b.effectivelyReachableWithBridging)  }
run reachable_and_ERWB_unordered for 1 but
   2 NetworkName, 2 NetworkState, 
   4 Name, 2 LinkIdent, 4 Link, 3 SessionIdent, 3 NetHdr,
   4 AcquireTable, 4 ForwardTable

/* --------------------------------------------------------------------
TESTS OF LAYERING
---------------------------------------------------------------------*/

pred LayeringTest1 [y: Layering] {
-- Fully one-connected network implemented on a hub-and-spoke network.
   let over = (y.overs).nets | let under = (y.unders).nets | {
      one over && one under
      Fully_oneLink_connected [over] && # over.members = 2
      Hub_and_spoke [under] && # under.members = 3
      over.links in y.oneImplemented   -- all overlay links implemented
      Underlay_satisfies_layering_demands [y]  
   }                                                                  }
run LayeringTest1 for 1 but 
   2 NetworkName, 2 NetworkState, 2 Bridging,
   3 Name, 4 LinkIdent, 6 Link, 2 SessionIdent, 2 NetHdr,
   5 AcquireTable, 5 ForwardTable, 4 ExternalReference

pred LayeringTest2 [y: Layering] {
-- Fully two-connected network implemented on a one-way ring.
   let over = (y.overs).nets | let under = (y.unders).nets | {
      one over && one under
      Fully_twoLink_connected [over] && # over.members = 3
      One_way_ring [under] && # under.members = 3
      over.links in y.twoImplemented  -- all overlay links implemented
      Underlay_satisfies_layering_demands [y]
   }                                                                 }
run LayeringTest2 for 1 but
   2 NetworkName, 2 NetworkState, 2 Bridging,
   3 Name, 2 LinkIdent, 9 Link, 3 SessionIdent, 6 NetHdr,
   6 AcquireTable, 6 ForwardTable, 6 ExternalReference

pred LayeringTest3 [y: Layering] {
-- One-way ring implemented on a spanning-tree network.
   let over = (y.overs).nets | let under = (y.unders).nets | {
      one over && one under
      One_way_ring [over] && # over.members = 3
      under in SpanningTreeState && # under.members = 4
      under.root ! in NetworkState.(Name.(y.attachments))
                           -- no overlay member is attached to the root
      over.links in y.oneImplemented   -- all overlay links implemented
      Underlay_satisfies_layering_demands [y]  
   }                                                                  }
run LayeringTest3 for 1 but 
   2 NetworkName, 2 NetworkState, 2 Bridging,
   4 Name, 3 LinkIdent, 9 Link, 3 SessionIdent, 6 NetHdr,
   7 AcquireTable, 7 ForwardTable, 6 ExternalReference

pred LayeringTest4 [y: Layering] {
-- Links are implemented on two bridged networks.  Reachability in
-- neither of these networks alone is sufficient.
   let over = (y.overs).nets | {
      one over 
      Fully_twoLink_connected [over] && # over.members = 2
      all disj m, m": over.members | 
         let n = m.(y.attachments).Name | 
         let n" = m".(y.attachments).Name |
         let a = n.(m.(y.attachments)) | 
         let a" = n".(m".(y.attachments)) | 
         (  some n && some n" && n != n"
         && (a -> a") ! in NetHdr.(n.reachable) 
         && (a -> a") ! in NetHdr.(n".reachable)  )
      over.links in y.twoImplemented  -- all overlay links implemented
      Underlay_satisfies_layering_demands [y]  
   }                                                                  }
run LayeringTest4 for 1 but 
   3 NetworkName, 3 NetworkState, 2 Bridging,
   4 Name, 2 LinkIdent, 8 Link, 1 SessionIdent, 2 NetHdr,
   6 AcquireTable, 6 ForwardTable, 4 ExternalReference

pred LayeringTest5 [y: Layering] {
-- Bridging links are implemented.
   let overs = (y.overs).nets | {
      # overs = 2
      some k, k": Link | 
      (  ReverseLink [k, k"] 
      && (k + k") in (y.implementations).SessionIdent  )
      Underlay_satisfies_layering_demands [y]  
   }                                                                  }
run LayeringTest5 for 1 but 
   3 NetworkName, 3 NetworkState, 2 Bridging,
   4 Name, 2 LinkIdent, 8 Link, 1 SessionIdent, 2 NetHdr,
   6 AcquireTable, 6 ForwardTable, 4 ExternalReference

pred LayeringTest6 [disj y, y": Layering] {
-- There can be a hierarchy of networks.  This validates 
-- Layering_is_hierarchical in particular.
   (y.unders).nets in (y".overs).nets 
   some y.implementations && Underlay_satisfies_layering_demands [y]  
   some y".implementations && Underlay_satisfies_layering_demands [y"]}
run LayeringTest6 for 1 but 
   3 NetworkName, 3 NetworkState, 2 Layering, 3 Bridging,
   3 Name, 2 LinkIdent, 6 Link, 3 SessionIdent, 3 NetHdr,
   7 AcquireTable, 7 ForwardTable, 6 ExternalReference

pred LayeringTest7 [disj y, y": Layering] {
-- Different layering objects can operate on the same underlays and
-- overlays.  This provides some validation of the four uniqueness
-- constraints.
   y.overs = y".overs && y.unders = y".unders
   some y.implementations
   some y".implementations
   Underlay_satisfies_layering_demands [y]  
   Underlay_satisfies_layering_demands [y"]  }  
run LayeringTest7 for 1 but 
   2 NetworkName, 2 NetworkState, 2 Layering, 2 Bridging,
   3 Name, 3 LinkIdent, 6 Link, 4 SessionIdent, 4 NetHdr,
   6 AcquireTable, 6 ForwardTable, 6 ExternalReference

pred LayeringTest8 [y: Layering] {
-- Self-links are implemented by self-sessions.
   some k: (y.implementations).SessionIdent | k.sndr = k.rcvrs
   Underlay_satisfies_layering_demands [y]                   }
run LayeringTest8 for 1 but 
   2 NetworkName, 2 NetworkState, 2 Bridging,
   4 Name, 2 LinkIdent, 8 Link, 1 SessionIdent, 2 NetHdr,
   6 AcquireTable, 6 ForwardTable, 4 ExternalReference

pred LayeringTest9 [disj y, y": Layering] {
-- Two layerings implement links of the same network on different 
-- networks.
   let over = (y.overs).nets | let over" = (y".overs).nets | {
      one over && one over"
      over != over"
      some y.implementations && some y".implementations
      Underlay_satisfies_layering_demands [y] 
      Underlay_satisfies_layering_demands [y"]                }  }
run LayeringTest9 for 1 but 
   3 NetworkName, 3 NetworkState, 3 Bridging, 2 Layering,
   6 Name, 2 LinkIdent, 8 Link, 2 SessionIdent, 2 NetHdr,
   6 AcquireTable, 6 ForwardTable, 4 ExternalReference

pred LayeringTest10 [disj y, y": Layering] {
-- Two layerings use one network to implement links of two different
-- networks.
   let over = (y.overs).nets | let over" = (y".overs).nets | {
      one over && one over" && one (y.unders).nets
      over != over"
      (y.unders).nets = (y".unders).nets
      some y.implementations && some y".implementations
      Underlay_satisfies_layering_demands [y] 
      Underlay_satisfies_layering_demands [y"]                }  }
run LayeringTest10 for 1 but 
   3 NetworkName, 3 NetworkState, 3 Bridging, 2 Layering,
   6 Name, 2 LinkIdent, 8 Link, 2 SessionIdent, 2 NetHdr,
   6 AcquireTable, 6 ForwardTable, 4 ExternalReference

pred LayeringTest11 [disj y, y": Layering, n: NetworkState, m: Name] {
-- An underlay member can be the attachment of members of more than
-- one overlay network.
   one (y.overs).nets && one (y".overs).nets
   (y.overs).nets != (y".overs).nets
   (n -> m) in Name.(y.attachments)
   (n -> m) in Name.(y".attachments)
   Underlay_satisfies_layering_demands [y]  
   Underlay_satisfies_layering_demands [y"]                          }
run LayeringTest11 for 1 but 
   3 NetworkName, 3 NetworkState, 2 Layering, 3 Bridging,
   4 Name, 3 LinkIdent, 6 Link, 2 SessionIdent, 2 NetHdr,
   6 AcquireTable, 6 ForwardTable, 4 ExternalReference

pred LayeringTest12 [disj y, y": Layering, m: Name] {
-- An overlay member can be attached to members of more than one
-- underlay network.
   y.overs = y".overs
   some disj n, n": NetworkState, disj a, a": Name |
   ((n -> a) in m.(y.attachments) && (n" -> a") in m.(y".attachments))
   Underlay_satisfies_layering_demands [y]  
   Underlay_satisfies_layering_demands [y"]                          }  
run LayeringTest12 for 1 but 
   3 NetworkName, 3 NetworkState, 2 Layering, 3 Bridging,
   4 Name, 3 LinkIdent, 6 Link, 2 SessionIdent, 2 NetHdr,
   6 AcquireTable, 6 ForwardTable, 4 ExternalReference

pred LayeringTest13 [y: Layering] {
-- Links of a non-local allGroup are implemented by MultiWay sessions.
   some g: ((y.overs).nets).allGroups, disj n, n": (y.overs).nets, 
        m: n.members, m": n".members, k: n.links & n".links |
   (  (m+m") in g.(n.groupReceivers) 
   && g.(n.groupReceivers) = g.(n".groupReceivers)  
   && g.(n.groupSenders) = g.(n".groupSenders)  
   && k.group = g
   && k in y.groupImplemented
   )
   Underlay_satisfies_layering_demands [y]                            }
run LayeringTest13 for 1 but 
   3 NetworkName, 3 NetworkState, 2 Bridging,
   3 Name, 3 LinkIdent, 6 Link, 1 SessionIdent, 1 NetHdr,
   6 AcquireTable, 6 ForwardTable, 3 ExternalReference

pred LayeringTest14 [y: Layering] {
-- Fully connecting allcast links are implemented by layering.
   let net = (y.overs).nets | {
      one net
      Fully_allCast_connected [net] 
      # net.members = 3
      net.groupLinks in y.groupImplemented
      Underlay_satisfies_layering_demands [y]            
}  }
run LayeringTest14 for 1 but 
   2 NetworkName, 2 NetworkState, 1 Layering, 2 Bridging,
   4 Name, 2 LinkIdent, 6 Link, 1 SessionIdent, 3 NetHdr,
   6 AcquireTable, 6 ForwardTable, 9 ExternalReference

pred LayeringTest15 [y: Layering] {
-- Fully connecting anycast links are implemented by layering.
   let net = (y.overs).nets | {
      one net
      Fully_anyCast_connected [net] 
      # net.members = 3
      # y.groupImplemented = 2     -- only groupLinks to be implemented
      Underlay_satisfies_layering_demands [y]  }  }
run LayeringTest15 for 1 but 
   2 NetworkName, 2 NetworkState, 1 Layering, 2 Bridging,
   6 Name, 3 LinkIdent, 9 Link, 2 SessionIdent, 2 NetHdr,
   6 AcquireTable, 6 ForwardTable, 4 ExternalReference

/* --------------------------------------------------------------------
TESTS OF SUBDUCTION
These declarations are necessary to run these tests:
one sig n1, n2 extends NetworkState {}
one sig m0, m1, m2, m3 extends Name {}
one sig k01, k12o, k12u, k23, k32 extends Link {}
-------------------------------------------------------------------- */

pred LittlestSubduction [s: Subduction] {
   no Link.group
   (s.overs.nets) = n1 && (s.unders).nets = n2
   s.sharedLinks = k01 + k23
   s.sharedMembers = m1 + m2
   k01.sndr = m0 && k01.rcvrs = m1
   k12o.sndr = m1 && k12o.rcvrs = m2
   k12u.sndr = m1 && k12u.rcvrs = m2
   k23.sndr = m2 && k23.rcvrs = m3
   (s.implementations).SessionIdent = k12o  }
run LittlestSubduction for 1 but
   2 NetworkName, 2 NetworkState, 2 Bridging, 
   4 Name, 2 LinkIdent, 5 Link, 2 SessionIdent, 2 NetHdr,
   6 AcquireTable, 6 ForwardTable, 2 ExternalReference

pred SubductionTest1 [s: SubductionWithTraffic] {
-- Previously unconnected underlay members reach each other through the
-- overlay.
   LittlestSubduction [s]
   n2.members = s.sharedMembers + m0 + m3
   NetHdr.((s.overs).reachableWithBridging) = (m1 -> m2)
   NetHdr.((s.unders).reachableWithBridging) = (m1 -> m2)
   (m0 -> m3) in NetHdr.(s.effectivelyReachableWithSubduction)  }
run SubductionTest1 for 1 but
   2 NetworkName, 2 NetworkState, 2 Bridging, 
   4 Name, 2 LinkIdent, 5 Link, 2 SessionIdent, 2 NetHdr,
   6 AcquireTable, 6 ForwardTable, 2 ExternalReference

pred SubductionTest2 [s: SubductionWithTraffic] {
-- The onramp shared link is external to the underlay network, but 
-- subduction is still useful.
   LittlestSubduction [s]
   n1.members = s.sharedMembers
   n2.members = s.sharedMembers + m3
   NetHdr.((s.overs).reachableWithBridging) = (m1 -> m2)
   NetHdr.((s.unders).reachableWithBridging) = (m1 -> m2)
   (m1 -> m3) in NetHdr.(s.effectivelyReachableWithSubduction)  }
run SubductionTest2 for 1 but
   2 NetworkName, 2 NetworkState, 2 Bridging, 
   4 Name, 2 LinkIdent, 5 Link, 2 SessionIdent, 2 NetHdr,
   6 AcquireTable, 6 ForwardTable, 2 ExternalReference

pred SubductionTest3 [s: SubductionWithTraffic] {
-- The offramp shared link is external to the underlay network, but 
-- subduction is still useful.
   LittlestSubduction [s]
   n1.members = s.sharedMembers
   n2.members = s.sharedMembers + m0
   NetHdr.((s.overs).reachableWithBridging) = (m1 -> m2)
   NetHdr.((s.unders).reachableWithBridging) = (m1 -> m2)
   (m0 -> m2) in NetHdr.(s.effectivelyReachableWithSubduction)  }
run SubductionTest3 for 1 but
   2 NetworkName, 2 NetworkState, 2 Bridging, 
   4 Name, 2 LinkIdent, 5 Link, 2 SessionIdent, 2 NetHdr,
   6 AcquireTable, 6 ForwardTable, 2 ExternalReference

pred SubductionTest4 [s: SubductionWithTraffic] {
-- One shared member ends both the onramp and the offramp, and this
-- overlay member forwards from on to off.
   no Link.group
   (s.overs.nets) = n1 && (s.unders).nets = n2
   s.sharedMembers = m2
   n1.members = s.sharedMembers + m1
   n2.members = s.sharedMembers + m0 + m3
   s.sharedLinks = k23 + k32
   k23.sndr = m2 && k23.rcvrs = m3
   k32.sndr = m3 && k32.rcvrs = m2
   k12o.sndr = m1 && k12o.rcvrs = m2
   k12u.sndr = m0 && k12u.rcvrs = m2
   (s.implementations).SessionIdent = k12o  
   NetHdr.((s.overs).reachableWithBridging) = (m1 -> m2)
   NetHdr.((s.unders).reachableWithBridging) = (m0 -> m2)
   (m3 -> m3) in NetHdr.(s.effectivelyReachableWithSubduction)  }
run SubductionTest4 for 1 but 
   2 NetworkName, 2 NetworkState, 2 Bridging, 
   4 Name, 2 LinkIdent, 5 Link, 2 SessionIdent, 2 NetHdr,
   5 AcquireTable, 5 ForwardTable, 2 ExternalReference

pred SubductionTest5 [s: Subduction] {
-- Two bridged overlay networks and two bridged underlay networks.
   # (s.overs).nets = 2
   # (s.unders).nets = 2
   some disj m, m": s.sharedMembers, disj o, o": (s.overs).nets,
      disj u, u": (s.unders).nets, k: (s.implementations).SessionIdent|
   (  m in o.members && m in u.members
   && m" in o".members && m" in u".members  
   && k.sndr = m && k.rcvrs = m"
   && k in (s.implementations).SessionIdent                           )
   some (s.overs+s.unders).reachableWithBridging
   some (s.reachableWithSubduction 
         - (s.overs+s.unders).reachableWithBridging)                  }
run SubductionTest5 for 1 but
   4 NetworkName, 4 NetworkState, 2 Bridging, 
   4 Name, 3 LinkIdent, 6 Link, 3 SessionIdent, 3 NetHdr,
   8 AcquireTable, 8 ForwardTable, 2 ExternalReference

/* --------------------------------------------------------------------
TESTS AND THEOREMS ABOUT THE PROPERTIES OF REACHABLE WITH SUBDUCTION
-------------------------------------------------------------------- */

-- According to the Bridging_closure_theorem, a bridging of networks is
-- equivalent to one network.  So this theorem is sufficient to capture
-- subduction semantics, in a way that is less operational than the
-- definition of reachableWithSubduction.
assert Subduction_semantics_theorem {
-- This theorem constructs, from the subduction of one network over
-- another, a third network.  It says that reachableWithSubduction is
-- equivalent to reachability in the third network.  The original two
-- networks cannot be bridged, because of their shared members.  Note
-- that the transmitTable of the constructed network does not matter,
-- because it plays no part in the computation of reachability.
   all s: Subduction, nn: NetworkState - (s.(overs+unders)).nets |
   let over = (s.overs).nets | let under = (s.unders).nets       |
   (  one over && one under && no Link.group
   && nn.members = (over + under).members 
   && nn.links = (over + under).links
   && (all m: nn.members |
         let oarows = (m.(over.acquireTables)).arows | 
         let uarows = (m.(under.acquireTables)).arows | 
         (m.(nn.acquireTables)).arows =
            { j: LinkIdent, h: NetHdr, c: AcquireCmd |
               (  ((j -> h -> c) in uarows && c in Receive + Forward)
               || (  (j -> h -> SharedAcquire) in uarows 
                  && (j -> h -> c) in oarows           )              
               )                                                    }
      )
   && (all m: nn.members |
         let ofrows = (m.(over.forwardTables)).frows | 
         let ufrows = (m.(under.forwardTables)).frows | 
         (m.(nn.forwardTables)).frows =
            { j: LinkIdent+NoInLink, h: NetHdr, c: LinkIdent+NoOutLink|
               (  ((j -> h -> c) in ofrows && c != SharedForward)
               || (  (j -> h -> SharedForward) in ofrows 
                  && (SharedPacket -> h -> c) in ufrows   )
               || ((j -> h -> c) in ufrows && j != SharedPacket)
               )                                                      }
      )
   ) => nn.reachable = s.reachableWithSubduction
}
check Subduction_semantics_theorem for 1 but
   2 NetworkName, 2 NetworkState, 2 Bridging, 
   4 Name, 2 LinkIdent, 5 Link, 2 SessionIdent, 2 NetHdr,
   5 AcquireTable, 5 ForwardTable, 2 ExternalReference

-- We do not need a Useful_RWS_over_RWB_test, because the previous
-- tests show that subduction adds to reachability.

assert Vestigial_RWS_over_RWB_theorem {
-- If there are no shared members, then subduction adds no reach.
   all s: Subduction | 
      no ((s.overs).nets).members & ((s.unders).nets).members
   => (s.overs + s.unders).reachableWithBridging 
         = s.reachableWithSubduction                        }
check Vestigial_RWS_over_RWB_theorem for 1 but
   2 NetworkName, 2 NetworkState, 2 Bridging, 
   4 Name, 2 LinkIdent, 5 Link, 2 SessionIdent, 2 NetHdr,
   5 AcquireTable, 5 ForwardTable, 2 ExternalReference

assert RWB_in_RWS_inclusion_theorem {
   all s: Subduction |
      (s.overs + s.unders).reachableWithBridging 
         in s.reachableWithSubduction          }
check RWB_in_RWS_inclusion_theorem for 1 but
   2 NetworkName, 2 NetworkState, 2 Bridging, 
   4 Name, 2 LinkIdent, 5 Link, 2 SessionIdent, 2 NetHdr,
   5 AcquireTable, 5 ForwardTable, 2 ExternalReference

pred Useful_RWS_over_ERWS_test [s: SubductionWithTraffic] {
-- This shows that reachableWithSubduction is bigger than its 
-- effective counterpart.
   some s.effectivelyReachableWithSubduction
   some (s.reachableWithSubduction 
         - s.effectivelyReachableWithSubduction)            }
run Useful_RWS_over_ERWS_test for 1 but
   2 NetworkName, 2 NetworkState, 2 Bridging, 
   4 Name, 2 LinkIdent, 5 Link, 2 SessionIdent, 2 NetHdr,
   5 AcquireTable, 5 ForwardTable, 2 ExternalReference

assert ERWS_in_RWS_inclusion_theorem {
   all s: SubductionWithTraffic |
      s.effectivelyReachableWithSubduction 
         in s.reachableWithSubduction                           }
check ERWS_in_RWS_inclusion_theorem for 1
but 2 NetworkName, 2 NetworkState, 
    5 Name, 3 LinkIdent, 5 Link, 1 SessionIdent, 1 NetHdr,
    5 AcquireTable, 5 ForwardTable

pred Useful_ERWS_over_ERWB_test [s: SubductionWithTraffic] {
   some (s.overs + s.unders).effectivelyReachableWithBridging
   some (s.effectivelyReachableWithSubduction
         - (s.overs+s.unders).effectivelyReachableWithBridging)   }
run Useful_ERWS_over_ERWB_test for 1 but
   2 NetworkName, 2 NetworkState, 2 Bridging, 
   4 Name, 2 LinkIdent, 5 Link, 2 SessionIdent, 2 NetHdr,
   5 AcquireTable, 5 ForwardTable, 2 ExternalReference
   
assert Vestigial_ERWS_over_ERWB_theorem {
-- If there are no shared members, then subduction adds no reach.
   all s: SubductionWithTraffic | 
      no ((s.overs).nets).members & ((s.unders).nets).members
   => (s.overs + s.unders).effectivelyReachableWithBridging 
         = s.effectivelyReachableWithSubduction                 }
check Vestigial_ERWS_over_ERWB_theorem for 1 but
   2 NetworkName, 2 NetworkState, 2 Bridging, 
   4 Name, 2 LinkIdent, 5 Link, 2 SessionIdent, 2 NetHdr,
   5 AcquireTable, 5 ForwardTable, 2 ExternalReference

assert ERWB_in_ERWS_inclusion_theorem {
   all s: SubductionWithTraffic |
      (s.overs + s.unders).effectivelyReachableWithBridging 
         in s.effectivelyReachableWithSubduction          }
check ERWB_in_ERWS_inclusion_theorem for 1 but
   2 NetworkName, 2 NetworkState, 2 Bridging, 
   4 Name, 2 LinkIdent, 5 Link, 2 SessionIdent, 2 NetHdr,
   5 AcquireTable, 5 ForwardTable, 2 ExternalReference

pred RWB_and_ERWS_unordered [s: SubductionWithTraffic] {
   some ((s.overs+s.unders).reachableWithBridging 
           - s.effectivelyReachableWithSubduction  )
   some (s.effectivelyReachableWithSubduction
           - (s.overs+s.unders).reachableWithBridging  )          }
run RWB_and_ERWS_unordered for 1 but
   2 NetworkName, 2 NetworkState, 2 Bridging, 
   4 Name, 2 LinkIdent, 5 Link, 2 SessionIdent, 2 NetHdr,
   5 AcquireTable, 5 ForwardTable, 2 ExternalReference
