/* ====================================================================
A MODEL OF GROUP COMMUNICATION	                           9 March 2023
==================================================================== */
open net[]
open traffic[]

/* --------------------------------------------------------------------
NETWORKS
   NetworkStateWithGroups adds group names, group links, and group 
sessions to the model.
-------------------------------------------------------------------- */

sig NetworkStateWithGroups in NetworkState {
   groups: set Name - members,                           -- group names
   disj allGroups, anyGroups: set groups,
   groupSenders: groups -> (Name - groups),
   groupReceivers: groups -> (Name - groups),
   groupLinks: set links,                                    -- derived
}  {
-- Constraints on groups.
   -- Each group is one of two types.
      allGroups + anyGroups = groups
   -- Consistency of group members.  Broadcast and multicast groups
   -- are conflated as AllCast groups.
      all g: groups | (some g.groupSenders && some g.groupReceivers)
      all g: anyGroups | no (g.groupSenders & g.groupReceivers)
-- Network topology.
   -- Derivation of link subsets.
      groupLinks = (links <: mode).(AllCast + AnyCast)
      oneLinks + twoLinks + groupLinks = links
   -- Consistency of link fields.
      all k: mode.AnyCast | (some k.group && k.group in anyGroups)
      all k: mode.AllCast | (some k.group && k.group in allGroups)
   -- A group link's sender and receivers match group and network
   -- membership.
      all k: mode.(AnyCast + AllCast), g: k.group | 
        (k.sndr in g.groupSenders && k.rcvrs = g.groupReceivers-k.sndr)  
   -- Note that, in the case of a member with allcast group links, 
   -- there can be multiple inLinks of the group with the same link 
   -- identifier, and also an outLink of the same group.  These are all
   -- links of an implementing set for the group.  Note that a group 
   -- can have more than one implementing set.
-- Behavior of network members.
   -- Reachability: Nondeterminism in a forwarding table means packet 
   -- replication, which should only be used for allCast sessions.
}

/* --------------------------------------------------------------------
PROPERTIES OF NETWORK TOPOLOGY
-------------------------------------------------------------------- */

pred Fully_allCast_connected [n: NetworkStateWithGroups] {
-- A network in which there is an allCast group, implemented by links,
-- including all members.
   some g: Name | {
   g in n.allGroups
   g.(n.groupSenders) = n.members 
   g.(n.groupReceivers) = n.members
   all m: n.members | some k: n.groupLinks | 
      (  k.group = g && k.sndr = m && k.rcvrs = n.members - m  )
   all m: n.members, disj k, k": n.links |
      (  k.group = g && k".group = g && m in k.sndr && m in k".rcvrs )
      => k.sndrIdent = m.(k".rcvrIdents)                           } }

pred AnyCast_connected_from_hub [n: NetworkStateWithGroups] {
-- A network in which there is an anyCast group, implemented by a link,
-- including all members except a distinguished sender.
   some g: Name, hub: n.members | {
   g in n.anyGroups
   g.(n.groupSenders) = hub 
   g.(n.groupReceivers) = n.members - hub
   some k: n.groupLinks | 
      (  k.group = g && k.sndr = hub && k.rcvrs = n.members - hub
      && one (n.members).(k.rcvrIdents)                         )  }  }

pred Fully_anyCast_connected [n: NetworkStateWithGroups] {
-- A network in which each member has both a unique name and a (unique)
-- anycast name.  There is an anycast link from each member to each
-- other member, with the receiver's anycast name as its group.  The
-- network is like an IP edge network using only private IP names.
   n in NetworkStateWithTraffic
   all m: n.members | one g: n.anyGroups |  
      (  g.(n.groupSenders) = n.members - m
      && g.(n.groupReceivers) = m         )
   all disj m, m": n.members | some k: n.links |
      (  k.sndr = m && k.rcvrs = m" && k.mode = AnyCast
      && k.group = (n.groupReceivers).m"              )  }

/* --------------------------------------------------------------------
PROPERTIES OF NETWORK TRAFFIC
-------------------------------------------------------------------- */

pred Fully_allCast_active [n: NetworkStateWithGroups] {
-- A network with an allCast session in which all members are two-way
-- participants.  Each receiver is forced to receive a distinct header 
-- from each sender, meaning that sending is authentic.
   n in NetworkStateWithTraffic
   some g: Name, s: SessionIdent | {
   g in n.allGroups
   g.(n.groupSenders) = n.members 
   g.(n.groupReceivers) = n.members
   all disj m, m": n.members | some h: NetHdr |
   (  h.src = m && h.dst = g && h.sessionIdent = s
   && (m -> h) in (n.effectiveSend).Link
   && (m" -> h) in (n.effectiveReceive).Link  )  }  }

pred Fully_anyCast_active [n: NetworkStateWithGroups] {
-- This network is like an IP edge network using only private IP names,
-- just like the network in Fully_anyCast_connected.  In this case,
-- however, there is an anycast session from each member to each other
-- member, in which the destination is reached by its anycast name.
   n in NetworkStateWithTraffic
   all m: n.members | one g: n.anyGroups |  
      (  g.(n.groupSenders) = n.members - m
      && g.(n.groupReceivers) = m         )
   all disj m, m": n.members | some h: NetHdr |
      (  h.src = m && h.dst = (n.groupReceivers).m"
      && (m -> h) in (n.effectiveSend).Link
      && (m" -> h) in (n.effectiveReceive).Link  )  
   all disj h, h": NetHdr | 
      (h.dst + h".dst) in n.anyGroups
   => h.sessionIdent != h".sessionIdent          }

/* --------------------------------------------------------------------
PROPERTIES OF NETWORK BEHAVIOR
-------------------------------------------------------------------- */

pred Has_allCast_service_through_forwarding [n:NetworkStateWithGroups]{
   Fully_allCast_active [n]
   Network_satisfies_communication_demands [n]
   no n.groupLinks                                                    }

pred Has_allCast_service_through_links [n:NetworkStateWithGroups] {
   Fully_allCast_active [n]
   Network_satisfies_communication_demands [n]
   Only_unicast_forwarding [n]             } -- there is no replication

pred Has_anyCast_service_through_forwarding [n:NetworkStateWithGroups]{
   Fully_anyCast_active [n]
   Network_satisfies_communication_demands [n]
   no n.groupLinks                                                    }

pred Has_anyCast_service_through_links [n:NetworkStateWithGroups] {
   Fully_anyCast_active [n]
   Network_satisfies_communication_demands [n]
   Fully_anyCast_connected [n]                                    }

/* ====================================================================
run P for 1 but
   N>=S NetworkName, S NetworkState,
   B Bridging, Y Layering, U Subduction,
   M Name, L<=K LinkIdent, K Link, D SessionIdent, D-2D Header,
   SxM AcquireTable, SxM ForwardTable, X ExternalReference
==================================================================== */
/* ====================================================================
VERIFICATION ARCHIVE
==================================================================== */   
/* --------------------------------------------------------------------
TESTS AND THEOREMS ABOUT NETWORK TOPOLOGY
-------------------------------------------------------------------- */

pred GroupTopologyTest1 [n: NetworkStateWithGroups] {
   Fully_allCast_connected [n]
   # n.members = 3                                  }
run GroupTopologyTest1 for 1 but
   4 Name, 3 Link, 3 AcquireTable, 3 ForwardTable

pred GroupTopologyTest2 [n: NetworkStateWithGroups] {
   Fully_allCast_connected [n] 
   No_external_links [n]
   No_self_links [n]                                }
run GroupTopologyTest2 for 1 but
   5 Name, 6 LinkIdent, 6 Link, 3 AcquireTable, 3 ForwardTable

pred GroupTopologyTest3 [n: NetworkStateWithGroups] {
   Fully_allCast_connected [n] 
   ! No_external_links [n]
   ! No_self_links [n]                              }
run GroupTopologyTest3 for 1 but
   5 Name, 6 LinkIdent, 6 Link, 3 AcquireTable, 3 ForwardTable

pred GroupTopologyTest4 [n: NetworkStateWithGroups] {
   Fully_allCast_connected [n] 
   Fully_twoLink_connected [n]                      }
run GroupTopologyTest4 for 1 but
   4 Name, 6 LinkIdent, 9 Link, 3 AcquireTable, 3 ForwardTable

pred GroupTopologyTest5 [n: NetworkStateWithGroups] {
   AnyCast_connected_from_hub [n] 
   # n.members = 3                                  }
run GroupTopologyTest5 for 1 but
   4 Name, 2 LinkIdent, 7 Link, 3 AcquireTable, 3 ForwardTable

pred GroupTopologyTest6 [n: NetworkStateWithGroups] {
   Fully_anyCast_connected [n] 
   # n.members = 3                                  }
run GroupTopologyTest6 for 1 but
   6 Name, 3 LinkIdent, 6 Link, 3 AcquireTable, 3 ForwardTable

assert GroupTopologyTheorem {
   all n: NetworkStateWithGroups |
      (  Fully_allCast_connected [n] && # n.members > 2  ) 
      => (  ! Point_to_point [n]
         && ! Hub_and_spoke [n]
         && ! One_way_ring [n]
         && n ! in SpanningTreeState  )                  }
check GroupTopologyTheorem for 1 but
   6 Name, 6 LinkIdent, 6 Link, 5 AcquireTable, 5 ForwardTable

/* --------------------------------------------------------------------
TESTS AND THEOREMS ABOUT NETWORK TRAFFIC
-------------------------------------------------------------------- */

pred GroupTrafficTest1 [n: NetworkStateWithGroups] {
      Fully_allCast_active [n] && Sending_is_authentic [n]
   && No_self_sessions [n] && # n.members = 3            }
run GroupTrafficTest1 for 1 but
   4 Name, 3 Link, 3 NetHdr, 3 AcquireTable, 3 ForwardTable

assert AllCastTrafficTheorem {
-- The definition of Session_identifiers_are_unique is designed for
-- point-to-point sessions only.  So it is not satisfied by allCast
-- sessions with more than two senders, even though they are correct.
   all n: NetworkStateWithGroups |
   (  Fully_allCast_active [n] && # n.members > 2  )
   => ! Session_identifiers_are_unique [n]                  }
check AllCastTrafficTheorem for 1 but
   4 Name, 3 Link, 3 SessionIdent, 3 NetHdr, 
   3 AcquireTable, 3 ForwardTable

pred GroupTrafficTest2 [n: NetworkStateWithGroups] {
      # n.members = 3 && Sending_is_authentic [n]
   && Fully_anyCast_active [n]                     }
run GroupTrafficTest2 for 1 but
   6 Name, 2 LinkIdent, 3 Link, 6 SessionIdent, 6 NetHdr, 
   3 AcquireTable, 3 ForwardTable

/* --------------------------------------------------------------------
TESTS OF NETWORK BEHAVIOR
-------------------------------------------------------------------- */

pred GroupBehaviorTest1 [n: NetworkStateWithGroups] {
   Has_allCast_service_through_forwarding [n]
   Only_authentic_traffic_delivered [n]
   Fully_reachable [n]
   # n.members = 3                                  }
run GroupBehaviorTest1 for 1 but
   4 Name, 3 LinkIdent, 3 Link, 3 NetHdr, 
   3 AcquireTable, 3 ForwardTable

pred GroupBehaviorTest2 [n: NetworkStateWithGroups] {
   Has_allCast_service_through_forwarding [n]
   Only_authentic_traffic_delivered [n]
   Fully_reachable [n]
   Fully_oneLink_connected [n]                      
   # n.members = 3                                  }
run GroupBehaviorTest2 for 1 but
   4 Name, 4 LinkIdent, 6 Link, 3 NetHdr, 
   3 AcquireTable, 3 ForwardTable

pred GroupBehaviorTest3 [n: NetworkStateWithGroups] {
   Has_allCast_service_through_forwarding [n]
   Only_authentic_traffic_delivered [n]
   Fully_reachable [n]
   One_way_ring [n]                      
   # n.members = 4                                  }
run GroupBehaviorTest3 for 1 but
   5 Name, 2 LinkIdent, 4 Link, 4 NetHdr, 
   4 AcquireTable, 4 ForwardTable

pred GroupBehaviorTest4 [n: NetworkStateWithGroups] {
   Has_allCast_service_through_forwarding [n]
   Only_authentic_traffic_delivered [n]
   Fully_reachable [n]
   n in SpanningTreeState              
   # n.members = 4                                  }
run GroupBehaviorTest4 for 1 but
   5 Name, 6 LinkIdent, 6 Link, 4 NetHdr, 
   4 AcquireTable, 4 ForwardTable

pred GroupBehaviorTest5 [n: NetworkStateWithGroups] {
   Has_allCast_service_through_links [n]
   Only_authentic_traffic_delivered [n]
   Fully_reachable [n]
   # n.members = 3                                  }
run GroupBehaviorTest5 for 1 but
   4 Name, 6 LinkIdent, 6 Link, 4 NetHdr, 
   3 AcquireTable, 3 ForwardTable

pred GroupBehaviorTest6 [n: NetworkStateWithGroups] {
   Has_anyCast_service_through_forwarding [n]
   # n.members = 3                                  }
run GroupBehaviorTest6 for 1 but
   6 Name, 6 LinkIdent, 6 Link, 6 SessionIdent, 6 NetHdr, 
   3 AcquireTable, 3 ForwardTable

pred GroupBehaviorTest7 [n: NetworkStateWithGroups] {
   Has_anyCast_service_through_links [n]
   # n.members = 3                                  }
run GroupBehaviorTest7 for 1 but
   6 Name, 4 LinkIdent, 6 Link, 6 SessionIdent, 6 NetHdr, 
   3 AcquireTable, 3 ForwardTable
