/* ====================================================================
A MODEL OF NETWORK STATE                                   3 March 2023
==================================================================== */
/* --------------------------------------------------------------------
TYPES
   Some instances of these types are not used here, but only in other
models based on this model.
-------------------------------------------------------------------- */

sig NetworkName { }
 
sig Name { }                            -- the name of a network member
   
sig LinkIdent, SessionIdent { }
 
one sig Primitive { }
abstract sig AcquireCmd { }
   one sig Receive, Forward, SharedAcquire extends AcquireCmd { }
abstract sig NoInLink { }
   one sig Self, SharedPacket extends NoInLink { }
abstract sig NoOutLink { }
   one sig Drop, SharedForward extends NoOutLink { }
abstract sig Mode { }                                  -- delivery mode
   one sig OneWay, TwoWay, AllCast, AnyCast extends Mode { }
abstract sig Protocol { }
   one sig TwoEndOneWay, TwoEndTwoWay, MultiWay extends Protocol { }

sig NetHdr
   {  src, dst: Name, sessionIdent: SessionIdent, 
      protocol: Protocol, overlay: lone NetworkName }
                         -- if the overlay is a distributed system that
                         -- is not a network, it need not be named
fact NetHdrsAreRecords { all h, h": NetHdr |
   (  h.sessionIdent = h".sessionIdent && h.overlay = h".overlay
   && h.src = h".src && h.dst = h".dst && h.protocol = h".protocol  ) 
   => h = h" }           -- if an object is a "record," the identity of
                        -- an instance depends only on its field values
pred ReverseNetHdr [h, h": NetHdr] {
      h.src = h".dst && h.dst = h".src && h.protocol = h".protocol
   && h.sessionIdent = h".sessionIdent && h.overlay = h".overlay }

sig Link { 
   sndr: Name, 
   sndrIdent: LinkIdent,
   rcvrs: some Name,
   rcvrIdents: rcvrs -> one LinkIdent,
   mode: Mode,
   group: lone Name 
}  {
   -- Consistency of group links.
   (some group) <=> mode in AllCast + AnyCast
   -- For a self-link, the mode and endpoint idents are constrained.
   sndr = rcvrs => (mode = OneWay && sndrIdent = rcvrs.(rcvrIdents))
}
-- For computation of reachability, a link must be directional.  So a
-- "link" represented in the model is actually a link plus direction;
-- a consequence of this is that a link can only have one sender. 
--    Note that links are not records; there can be two distinct
-- individuals that happen to have the same field values.
pred ReverseLink [k, k": Link] {
      no k.group && no k".group
   && k.sndr = k".rcvrs && k.sndrIdent = (k".rcvrs).(k".rcvrIdents)
   && k.rcvrs = k".sndr && (k.rcvrs).(k.rcvrIdents) = k".sndrIdent  }
   
sig ExternalReference { exNet: NetworkName }
sig ExternalLink extends ExternalReference { exLink: LinkIdent }
sig ExternalSession extends ExternalReference {exSession: SessionIdent}
 
sig AcquireTable { arows: LinkIdent -> NetHdr -> AcquireCmd }

sig ForwardTable { frows:
   (LinkIdent + NoInLink) -> NetHdr -> (LinkIdent + NoOutLink) }

/* --------------------------------------------------------------------
NETWORK STATES
-------------------------------------------------------------------- */

-- All networks have unique names.
fact { networkName in NetworkState lone -> NetworkName }
   
sig NetworkState {
-- Network topology.
   networkName: NetworkName,
   members: set Name,        -- every member has at least a unique name
   disj infras, users: set members,    -- trusted and untrusted members
   links: set Link,
   disj oneLinks, twoLinks: set links,                       -- derived
   outLinks, inLinks: members -> LinkIdent -> links,         -- derived
-- Behavior of network members.
   acquireTables: members -> lone AcquireTable,
   forwardTables: members -> lone ForwardTable,
   transmitTable: 
      members -> LinkIdent -> (Primitive + ExternalSession),
   oneStep: NetHdr -> links -> links,                        -- derived
   reachable: NetHdr -> members -> members                   -- derived
}  {
-- Network topology.
   -- Each member is an infrastructure member or user.
      infras + users = members
   -- Derivation of link subsets.
      oneLinks = (links <: mode).OneWay
      twoLinks = (links <: mode).TwoWay
   -- All links have at least one endpoint that is a member.
      all k: links | some ((k.sndr + k.rcvrs) & members) 
   -- Point-to-point link fields have the right types.
      all k: oneLinks + twoLinks | (one k.rcvrs && no k.group)
      all k: twoLinks | some k": twoLinks - k | ReverseLink [k, k"]
   -- At each member, all outgoing links have distinct identifiers.
      all disj k, k": links | 
         k.sndr = k".sndr => k.sndrIdent != k".sndrIdent
   -- At each member, all incoming links have distinct identifiers, 
   -- unless they are links of the same group with different senders.
      all disj k, k": links, m: (k.rcvrs & k".rcvrs)|
      (  m.(k.rcvrIdents) != m.(k".rcvrIdents)
      || (k.group = k".group && k.sndr != k".sndr)  )
   -- At each member, all incoming links and outgoing links have
   -- distinct identifiers, unless they are reverse links or links
   -- implementing the same group.
      all disj k, k": links, m: (k.sndr & k".rcvrs)|
      (  k.sndrIdent != m.(k".rcvrIdents)
      || ReverseLink [k, k"]
      || some k.group && k.group = k".group        )
   -- At a member, reverse links must share a link identifier.
      all k, k": links | ReverseLink [k, k"] 
      => k.sndrIdent = (k".rcvrs).(k".rcvrIdents) 
   -- Derivation of inLinks and outLinks. 
      outLinks = { m: members, i: LinkIdent, k: links |
         k.sndr = m && k.sndrIdent = i                }         
      inLinks = { m: members, i: LinkIdent, k: links |
         (m -> i) in k.rcvrIdents                    }
-- Behavior of network members.  
-- Acquire and forward tables have to be modeled differently because 
-- they have too many columns for Alloy.  
   -- The acquireTable.
      -- Consistency:  all inLinks are real.
      all m: members | 
         (((m.acquireTables).arows).AcquireCmd).NetHdr
      in (m.inLinks).links
      -- An acquireTable can yield both Receive and Forward because, to
      -- implement allCast service, a member may have to both receive
      -- and forward a packet.  Otherwise it is deterministic.  If 
      -- there is no matching entry for an incoming packet, it is
      -- dropped.
         all m: members, ki: LinkIdent, h: NetHdr |
            (  lone h.(ki.((m.acquireTables).arows))
            || h.(ki.((m.acquireTables).arows)) = Receive + Forward  )
   -- The forwardTable.
      -- Consistency: All inLinks and outLinks are real.
      all m: members, kin: 
         (((m.forwardTables).frows).(LinkIdent + NoOutLink)).NetHdr |
         ( (m -> kin) in inLinks.links || kin in NoInLink )
      all m: members, kout:
              NetHdr.((LinkIdent+NoInLink).((m.forwardTables).frows)) |
         kout ! in NoOutLink => (m -> kout) in outLinks.links
      -- A forwardTable can yield multiple outLinks for replication for
      -- allCast service.  Otherwise it is deterministic.
         all m: members, ki: LinkIdent + NoOutLink, h: NetHdr |
            (  lone h.(ki.((m.forwardTables).frows))
            || h.(ki.((m.forwardTables).frows)) in LinkIdent  )
   -- The transmitTable.  
      -- The transmitTable is complete with respect to the forwarding
      -- table.
         { m: members, ki: LinkIdent | ki in
            NetHdr.((LinkIdent + NoInLink).((m.forwardTables).frows)) }
               in transmitTable.(Primitive+ExternalSession)
      -- The transmitTable is deterministic, because a link has only
      -- one embodiment.
         all m: members, ki: LinkIdent | lone ki.(m.transmitTable)
-- Reachability.
      oneStep = {  h: NetHdr, disj k, k": links |
         some m: members, kin, kout: LinkIdent |
            (m -> kin -> k) in inLinks 
         && (m -> kout -> k") in outLinks
         && (kin -> h -> Forward) in (m.acquireTables).arows
         && (kin -> h -> kout) in (m.forwardTables).frows  }
      reachable = {  h: NetHdr, m, m": members | 
         some k, k": links, kin, kout: LinkIdent | 
            (m -> kout -> k) in outLinks
         && (m" -> kin -> k") in inLinks
         && (h -> kout) in 
               (LinkIdent + Self).((m.forwardTables).frows)
         && (kin -> h) in ((m".acquireTables).arows).(Receive+Forward)
         && (  k = k" || (k -> k") in ^(h.oneStep)  )                }
}

/* --------------------------------------------------------------------
PROPERTIES OF NETWORK TOPOLOGY
-------------------------------------------------------------------- */

pred Point_to_point [n: NetworkState] {
-- All the links in the network are point-to-point links.
   n.links in (n.oneLinks + n.twoLinks)  }

pred No_external_links [n: NetworkState] {
   all k: n.links | (k.sndr + k.rcvrs) in n.members  }

pred No_self_links [n: NetworkState] {
   all k: n.links | k.sndr ! in k.rcvrs  }

pred Fully_oneLink_connected [n: NetworkState] {
-- A network in which there is a one-way link from each member to each
-- other member.  It does not matter whether the members are infras or
-- users, because no forwarding is necessary.
   all disj m, m": n.members | some k: n.oneLinks |
      k.sndr = m && k.rcvrs = m"                  }

pred Fully_twoLink_connected [n: NetworkState] {
-- A network in which there is a direct two-way link between each pair 
-- of members.  It does not matter whether the members are infras or
-- users, because no forwarding is necessary.
   all disj m, m": n.members | some k: n.twoLinks |
      k.sndr = m && k.rcvrs = m"                  }

pred Hub_and_spoke [n: NetworkState] {
   -- There is one forwarder, at the hub.
   one n.infras
   -- Users are connected to the forwarder by twoLinks.
   all m: n.infras, m": n.users |
      some k, k": n.twoLinks |
         (  k.sndr = m && k.rcvrs = m" && ReverseLink [k, k"]  )
   -- Users are not directly connected to each other.
   No_self_links [n]
   no k: n.links | k.sndr + k.rcvrs in n.users                 }

pred One_way_ring [n: NetworkState] {
   no n.users && n.links in n.oneLinks 
   No_external_links [n] && No_self_links [n]
   -- One-way links form a ring.
   let connections = { disj m0, m1: n.members |
      some k: n.links | m0 = k.sndr && m1 = k.rcvrs } |
   (  connections in n.members one -> one n.members 
   && all m, m": n.members | (m -> m") in ^connections  )  }

sig SpanningTreeState in NetworkState {
   root: members,
   disj downLinks, upLinks: set twoLinks,
   downGraph: members -> members           }                 -- derived
{
   links = downLinks + upLinks
   some members - root
-- All links are in directional pairs.
   all dk: downLinks | one uk: upLinks | ReverseLink [dk, uk]
   all uk: upLinks | one dk: downLinks | ReverseLink [uk, dk]
-- Derivation of the downGraph.
   downGraph = {  up: members, down: members |
      some dk: downLinks | up = dk.sndr && down = dk.rcvrs  }
-- The graph of downLinks is a tree.
   all m: members - root | m in root.(^downGraph)
   no m: members | m in m.(^downGraph)
   all m: members - root | one downGraph.m
-- Note that a spanning-tree network can have external links that are
-- not in the graph.
}

/* --------------------------------------------------------------------
PROPERTIES OF NETWORK BEHAVIOR
-------------------------------------------------------------------- */

pred No_routing_loops [n: NetworkState] {
   all h: NetHdr | no k: n.links | (k -> k) in ^(h.(n.oneStep)) }

pred Fully_reachable [n: NetworkState] {
-- Every member can reach every other member.
   all disj m0, m1: n.members |
      some h: NetHdr | (h -> m0 -> m1) in n.reachable  }

pred Only_unicast_forwarding [n: NetworkState] {
-- There is no group (broadcast or multicast) forwarding.
   all m: n.members, ki: LinkIdent + NoInLink, h: NetHdr |
      lone h.(ki.(m.((n.forwardTables).frows)))          }

/* --------------------------------------------------------------------
NETWORK EQUIVALENCE
-------------------------------------------------------------------- */

pred Networks_equivalent [n, n": NetworkState] {
-- This is the fundamental comparison operator on networks.  Theorems
-- show that it is reflexive, symmetric, and transitive, hence an
-- equivalence relation.
   n.reachable = n".reachable                  }
  
/* ====================================================================
run P for 1 but
   N>=S NetworkName, S NetworkState,
   M Name, L<=K LinkIdent, K Link, D SessionIdent, D-2D NetHdr,
   P Protocol,
   SxM AcquireTable, SxM ForwardTable, X ExternalReference
==================================================================== */
/* ====================================================================
VALIDATION AND VERIFICATION ARCHIVE
==================================================================== */
/* --------------------------------------------------------------------
TESTS OF NETWORK TOPOLOGY
-------------------------------------------------------------------- */

pred TopologyTest1 [n: NetworkState] { 
   some n.links && No_external_links [n]  }
run TopologyTest1 for 1 but
   3 Name, 2 LinkIdent, 6 Link, 3 AcquireTable, 3 ForwardTable

pred TopologyTest2 [n: NetworkState] { 
   # n.members = 3 && Point_to_point [n]  }
run TopologyTest2 for 1 but
   3 Name, 4 LinkIdent, 6 Link, 3 AcquireTable, 3 ForwardTable

pred TopologyTest3 [n: NetworkState] { 
   # n.members = 3 && Fully_oneLink_connected[n]  }
run TopologyTest3 for 1 but
   3 Name, 4 LinkIdent, 6 Link, 3 AcquireTable, 3 ForwardTable

pred TopologyTest4 [n: NetworkState] { 
   # n.members = 3 && Fully_twoLink_connected[n]  }
run TopologyTest4 for 1 but
   3 Name, 2 LinkIdent, 6 Link, 3 AcquireTable, 3 ForwardTable

pred TopologyTest5 [n: NetworkState] { 
   # n.members = 4 && Hub_and_spoke [n]  }
run TopologyTest5 for 1 but
   4 Name, 3 LinkIdent, 6 Link, 4 AcquireTable, 4 ForwardTable

pred TopologyTest6 [n: NetworkState] { 
   # n.members = 4 && One_way_ring [n]  }
run TopologyTest6 for 1 but
   4 Name, 2 LinkIdent, 4 Link, 4 AcquireTable, 4 ForwardTable

pred TopologyTest7 [n: SpanningTreeState] {
   # n.members = 5 && ! Hub_and_spoke [n]  }
run TopologyTest7 for 1 but
   5 Name, 4 LinkIdent, 8 Link, 5 AcquireTable, 5 ForwardTable

pred TopologyTest8 [n: SpanningTreeState] {
      # n.members = 5 && Hub_and_spoke [n] && No_external_links [n]  }
run TopologyTest8 for 1 but
   5 Name, 4 LinkIdent, 8 Link, 5 AcquireTable, 5 ForwardTable

/* --------------------------------------------------------------------
TESTS AND THEOREMS ABOUT NETWORK BEHAVIOR
-------------------------------------------------------------------- */

pred BehaviorTest1 [n: NetworkState] { ! No_routing_loops [n] }
run BehaviorTest1 for 1 but
   3 Name, 3 LinkIdent, 6 Link, 3 AcquireTable, 3 ForwardTable

pred BehaviorTest2 [n: NetworkState] { ! Fully_reachable [n] }
run BehaviorTest2 for 1 but
   3 Name, 3 LinkIdent, 6 Link, 3 AcquireTable, 3 ForwardTable

pred BehaviorTest3 [n: NetworkState] { ! Only_unicast_forwarding [n] }
run BehaviorTest3 for 1 but
   3 Name, 3 LinkIdent, 6 Link, 3 AcquireTable, 3 ForwardTable

pred BehaviorTest4 [n: NetworkState] { 
      No_routing_loops [n] && Fully_oneLink_connected [n]  }
run BehaviorTest4 for 1 but
   3 Name, 3 LinkIdent, 6 Link, 3 AcquireTable, 3 ForwardTable

pred BehaviorTest5 [n: NetworkState] { 
      Fully_reachable [n] && One_way_ring[n]  }
run BehaviorTest5 for 1 but
   4 Name, 2 LinkIdent, 4 Link, 4 AcquireTable, 4 ForwardTable

pred BehaviorTest6 [n: NetworkState] {
      Only_unicast_forwarding [n] && Fully_twoLink_connected [n]  }
run BehaviorTest6 for 1 but
   3 Name, 3 LinkIdent, 6 Link, 3 AcquireTable, 3 ForwardTable

pred BehaviorTest7 [n: SpanningTreeState] {
   No_routing_loops [n] && Only_unicast_forwarding [n] &&
   Fully_reachable [n] && ! No_external_links [n] }
run BehaviorTest7 for 1 but
   4 Name, 3 LinkIdent, 8 Link, 4 NetHdr,
   4 AcquireTable, 4 ForwardTable

pred BehaviorTest8 [n: NetworkState] {
      Only_unicast_forwarding [n] && ! Point_to_point [n]  }
run BehaviorTest8 for 1 but
   3 Name, 3 LinkIdent, 6 Link, 3 AcquireTable, 3 ForwardTable

assert Hub_and_spoke_nonreachability_theorem {
-- In a hub-and-spoke network, if the user members do not forward and
-- the hub does not send or receive, then the hub should not be 
-- reachable from itself.
   all n: NetworkState | {
      Hub_and_spoke [n]
      all f: n.infras |
         let atbl = ((f.(n.acquireTables)).arows) |
         let ftbl = ((f.(n.forwardTables)).frows) |
      (  -- The only acquire command is Forward.
         NetHdr.(LinkIdent.atbl) = Forward
         -- Only packets from inLinks are forwarded.
      && (ftbl.(LinkIdent+NoOutLink)).NetHdr in LinkIdent  )
      all u: n.users |
         let atbl = ((u.(n.acquireTables)).arows) |
         let ftbl = ((u.(n.forwardTables)).frows) |
      (  -- The only acquire command is Receive. 
         NetHdr.(LinkIdent.atbl) = Receive
         -- Only sent packets are forwarded.
      && (ftbl.(LinkIdent+NoOutLink)).NetHdr in Self  )
   }  => all f: n.infras | (f -> f) ! in NetHdr.(n.reachable)  }
check Hub_and_spoke_nonreachability_theorem for 1 but
   5 Name, 4 LinkIdent, 8 Link, 2 SessionIdent, 2 NetHdr,
   5 AcquireTable, 5 ForwardTable

assert One_way_ring_theorem {
-- If members receive packets with their own name as destination and 
-- forward other packets from their inLinks to their outLinks, then 
-- all members are reachable from all other members.  Note that this
-- formulation of the theorem does not work if the network has
-- external links.
   all n: NetworkState |
   (  One_way_ring [n] 
   && all m: n.members, h: NetHdr |
      h.dst = m =>
      (  (let kin = (m.(n.inLinks)).Link |
         let kout = (m.(n.outLinks)).Link |
         let atbl = ((m.(n.acquireTables)).arows) |
         let ftbl = ((m.(n.forwardTables)).frows) |
            h.(kin.atbl) = Receive && (Self -> h -> kout) in ftbl  )
      && (all m": n.members - m | 
            let kin = (m".(n.inLinks)).Link |
            let kout = (m".(n.outLinks)).Link |
            let atbl = ((m".(n.acquireTables)).arows) |
            let ftbl = ((m".(n.forwardTables)).frows) |
               h.(kin.atbl) = Forward && (kin -> h -> kout) in ftbl  )
      )
   )  => (  all m, m": n.members, h: NetHdr |
               h.dst = m" => (h -> m -> m") in n.reachable  )        }
check One_way_ring_theorem for 1 but
   5 Name, 4 LinkIdent, 8 Link, 5 AcquireTable, 5 ForwardTable

/* --------------------------------------------------------------------
TESTS AND THEOREMS ABOUT EQUIVALENCE
-------------------------------------------------------------------- */

pred EquivalenceTest1 [disj n, n": NetworkState] { 
   Networks_equivalent [n, n"] && n.links != n".links  }
run EquivalenceTest1 for 1 but 2 NetworkName, 2 NetworkState,
   4 Name, 3 LinkIdent, 8 Link, 4 AcquireTable, 4 ForwardTable

pred EquivalenceTest2 [disj n, n": NetworkState] { 
   Networks_equivalent [n, n"] && n.members != n".members  }
run EquivalenceTest2 for 1 but 2 NetworkName, 2 NetworkState,
   4 Name, 3 LinkIdent, 8 Link, 4 AcquireTable, 4 ForwardTable

assert Equivalence_reflexivity_theorem {
   all n: NetworkState | Networks_equivalent [n, n]  }
check Equivalence_reflexivity_theorem for 1 but
   6 Name, 4 LinkIdent, 8 Link, 2 SessionIdent, 2 NetHdr,
   6 AcquireTable, 6 ForwardTable

assert Equivalence_symmetry_theorem {
   all n, n": NetworkState | 
      Networks_equivalent [n, n"] => Networks_equivalent [n", n]  }
check Equivalence_symmetry_theorem for 1 but
   2 NetworkName, 2 NetworkState,
   6 Name, 4 LinkIdent, 8 Link, 2 SessionIdent, 2 NetHdr,
   6 AcquireTable, 6 ForwardTable

assert Equivalence_transitivity_theorem {
   all n0, n1, n2: NetworkState | 
      Networks_equivalent [n0, n1] && Networks_equivalent [n1, n2]
   => Networks_equivalent [n0, n2]                               }
check Equivalence_transitivity_theorem for 1 but
   3 NetworkName, 3 NetworkState,
   6 Name, 4 LinkIdent, 8 Link, 2 SessionIdent, 2 NetHdr,
   6 AcquireTable, 6 ForwardTable
