/* ====================================================================
A MODEL OF EXPECTED NETWORK TRAFFIC                        3 March 2023
==================================================================== */
open net[]
   
/* --------------------------------------------------------------------
NETWORKS
-------------------------------------------------------------------- */

sig NetworkStateWithTraffic in NetworkState {
-- Network traffic.
   sendTable: members -> NetHdr -> SessionIdent,
   receiveTable:
      members -> NetHdr -> (Primitive + ExternalLink),
   effectiveSend: members -> NetHdr -> Link,                 -- derived
   effectiveReceive: members -> NetHdr -> Link,              -- derived
-- Behavior of network members.
   effectivelyReachable: NetHdr -> members -> members        -- derived
}  {
-- Network traffic.  This models the traffic that members are expected
-- to send and receive, or that the network is required to carry, and 
-- it is optional in the model of an individual network.  Note that 
-- infrastructure members can send and receive.
   -- Determinism and consistency of the sendTable.
      all m: members, s: SessionIdent | let h = (m.sendTable).s |
         (  lone h && (some h => h.sessionIdent = s)  )
   -- Determinism of the receiveTable.
      all m: members, h: NetHdr | lone h.(m.receiveTable)
-- Derivations of effectiveSend and effectiveReceive.  
-- If a send is expected, it should be supported by an entry in the 
-- member's forwardTable.  If a receive is expected, it is supported by
-- an entry in the member's acquireTable.
      effectiveSend = { m: members, h: NetHdr, k: Link |
            (m -> h) in sendTable.SessionIdent
         && some ki: LinkIdent |
            (  (m -> ki -> k) in outLinks                
            && (h -> ki) in Self.((m.forwardTables).frows)  ) }
      effectiveReceive = { m: members, h: NetHdr, k: Link |
            (m -> h) in receiveTable.(Primitive+ExternalLink)
         && some ki: LinkIdent |
            (  (m -> ki -> k) in inLinks                    
            && (ki -> h -> Receive) in (m.acquireTables).arows  ) }
-- Effective reachability.
-- Note that the definitions of expectedSend and expectedReceive
-- subsume lines 5-7 of the definition of reachable.
      effectivelyReachable = {  h: NetHdr, m, m": members |
         some k, k": links |
            (m -> h -> k) in effectiveSend
         && (m" -> h -> k") in effectiveReceive
         && (  k = k" || (k -> k") in ^(h.oneStep)  )      }
}

/* --------------------------------------------------------------------
PROPERTIES OF NETWORK TRAFFIC
-------------------------------------------------------------------- */

pred Fully_oneSession_active [n: NetworkStateWithTraffic] {
-- A network with a one-way session from each member to each other
-- member.
   all disj m, m": n.members | some h: NetHdr |
      h.src = m && (m -> h) in (n.effectiveSend).Link
   && h.dst = m" && (m" -> h) in (n.effectiveReceive).Link  }

pred Users_fully_oneSession_active [n: NetworkStateWithTraffic] {
-- A network with a one-way session from each user member to each other
-- user member.
   all disj m, m": n.users | some h: NetHdr |
      h.src = m && (m -> h) in (n.effectiveSend).Link
   && h.dst = m" && (m" -> h) in (n.effectiveReceive).Link  }

pred Fully_twoSession_active [n: NetworkStateWithTraffic] {
-- A network with a two-way session between each pair of members.
   all disj m, m": n.members | some h, h": NetHdr |
   (  ReverseNetHdr [h, h"]
   && h.src = m && h.dst = m"
   && (m -> h) in (n.effectiveSend).Link
   && (m" -> h") in (n.effectiveSend).Link
   && (m" -> h) in (n.effectiveReceive).Link
   && (m -> h") in (n.effectiveReceive).Link  )  }

pred Users_fully_twoSession_active [n: NetworkStateWithTraffic] {
-- A network with a two-way session between each pair of user members.
   all disj m, m": n.users | some h, h": NetHdr |
   (  ReverseNetHdr [h, h"]
   && h.src = m && h.dst = m"
   && (m -> h) in (n.effectiveSend).Link
   && (m" -> h") in (n.effectiveSend).Link
   && (m" -> h) in (n.effectiveReceive).Link
   && (m -> h") in (n.effectiveReceive).Link  )  }

pred No_self_sessions [n: NetworkStateWithTraffic] {
-- Enabling a member to have a session with itself is not required.
   all m: n.members |
      no (m.(n.effectiveSend).Link & m.(n.effectiveReceive).Link)  }

pred Sending_is_authentic [n: NetworkStateWithTraffic] {
-- Members only send packets with their own name in the source field.
   all m: n.members, h: NetHdr |
      (m -> h) in (n.sendTable).SessionIdent => h.src = m  }

pred Receiving_is_authentic [n: NetworkStateWithTraffic] {
-- Members only receive packets with their own name in the destination
-- field.
   all m: n.members, h: NetHdr |
      (m -> h) in (n.receiveTable).(Primitive+ExternalLink)
    => h.dst = m  }

pred Session_identifiers_are_unique [n: NetworkStateWithTraffic] {
-- If two different headers have the same session identifier, then
-- the headers reverse each other.
   all s: NetHdr.((n.members).(n.sendTable)) |
      all disj h, h": NetHdr |
         (h -> s) + (h" -> s) in (n.members).(n.sendTable)
      => ReverseNetHdr [h, h"]                                   }

pred Traffic_is_symmetric [n: NetworkStateWithTraffic] {
-- The traffic model only includes well-formed two-way sessions.
   all h: NetHdr, m, m": n.members |
   (  (m -> h) in (n.sendTable).SessionIdent
   && (m" -> h) in (n.receiveTable).(Primitive+ExternalLink)  )
   => some h": NetHdr | 
      (  ReverseNetHdr [h, h"]
      && (m" -> h") in (n.sendTable).SessionIdent
      && (m -> h") in (n.receiveTable).(Primitive+ExternalLink)  )  }

/* --------------------------------------------------------------------
PROPERTIES OF NETWORK BEHAVIOR
-------------------------------------------------------------------- */

pred Network_satisfies_communication_demands 
   [n: NetworkStateWithTraffic]              {
-- Communication demands are specified by the send and receive tables.
   all m, m": n.members, h: NetHdr |
   (  (m -> h) in (n.sendTable).SessionIdent
   && (m" -> h) in (n.receiveTable).(Primitive+ExternalLink)  )
   => (h -> m -> m") in n.effectivelyReachable                }

pred Only_authentic_traffic_delivered [n: NetworkStateWithTraffic] {
   all m, m": n.members, h: NetHdr |
      (h -> m -> m") in n.effectivelyReachable => h.src = m        }

pred Effective_reachability_is_symmetric [n: NetworkStateWithTraffic] {
   all h: NetHdr, m, m": n.members |
      (h -> m -> m") in n.effectivelyReachable
   => some h": NetHdr | ReverseNetHdr [h, h"]
      && (h" -> m" -> m) in n.effectivelyReachable                    }

/* --------------------------------------------------------------------
NETWORK EQUIVALENCE
-------------------------------------------------------------------- */

pred Networks_effectively_equivalent [n, n": NetworkState] {
-- This is the fundamental comparison operator on networks.  Theorems
-- show that it is reflexive, symmetric, and transitive, hence an
-- equivalence relation.
   n.effectivelyReachable = n".effectivelyReachable                  }

/* ====================================================================
run P for 1 but
   N>=S NetworkName, S NetworkState,
   M Name, L<=K LinkIdent, K Link, D SessionIdent, D-2D NetHdr,
   SxM AcquireTable, SxM ForwardTable, X ExternalReference
==================================================================== */
/* ====================================================================
VALIDATION AND VERIFICATION ARCHIVE
==================================================================== */
/* --------------------------------------------------------------------
TESTS AND THEOREMS ABOUT EFFECTIVE REACHABILITY
-------------------------------------------------------------------- */
 
pred Useful_effectiveness_test [n: NetworkStateWithTraffic] {
-- Effectiveness constrains reachability.
   some n.effectivelyReachable 
   some n.reachable - n.effectivelyReachable                }
run Useful_effectiveness_test for 1 but
   5 Name, 3 LinkIdent, 6 Link, 5 AcquireTable, 5 ForwardTable
 
assert Effective_reachability_inclusion_theorem {
   all n: NetworkStateWithTraffic | 
      n.effectivelyReachable in n.reachable  }
check Effective_reachability_inclusion_theorem for 1 but
   5 Name, 3 LinkIdent, 6 Link, 5 AcquireTable, 5 ForwardTable

pred Effective_reach_excludes_middlebox [n: NetworkStateWithTraffic] {
   Point_to_point [n] && Only_unicast_forwarding [n]
   some h: NetHdr, disj m0, m1, m2: n.members |
      (h ->m0 ->m1) + (h ->m0 ->m2) + (h ->m1 ->m2) in n.reachable
   && (h ->m0 ->m2) in n.effectivelyReachable
   && (h -> m0 -> m1) ! in n.effectivelyReachable                
   && (h -> m1 -> m2) ! in n.effectivelyReachable                     }
run Effective_reach_excludes_middlebox for 1 but
   3 Name, 2 LinkIdent, 4 Link, 3 AcquireTable, 3 ForwardTable

/* --------------------------------------------------------------------
TESTS OF NETWORK TRAFFIC
-------------------------------------------------------------------- */

pred TrafficTest1 [n: NetworkStateWithTraffic] {  
   Fully_oneSession_active [n]
   No_self_sessions [n]              
   # n.members = 4 && One_way_ring [n]         }
run TrafficTest1 for 1 but
   4 Name, 2 LinkIdent, 6 Link, 12 SessionIdent, 12 NetHdr,
   4 AcquireTable, 4 ForwardTable

pred TrafficTest2 [n: NetworkStateWithTraffic] {   
   # n.members = 4 
   Hub_and_spoke [n]              
   Users_fully_oneSession_active [n]  
   Sending_is_authentic [n]
   Receiving_is_authentic [n]
}
run TrafficTest2 for 1 but
   4 Name, 3 LinkIdent, 6 Link, 6 SessionIdent, 6 NetHdr,
   4 AcquireTable, 4 ForwardTable

pred TrafficTest3 [n: NetworkStateWithTraffic] {  
   Fully_twoSession_active [n]
   Session_identifiers_are_unique [n]
   # n.members = 4 && One_way_ring [n]              }
run TrafficTest3 for 1 but
   4 Name, 2 LinkIdent, 4 Link, 6 SessionIdent, 12 NetHdr,
   4 AcquireTable, 4 ForwardTable

pred TrafficTest4 [n: NetworkStateWithTraffic] { 
   n in SpanningTreeState
   ! Hub_and_spoke [n]
   some n.infras && some n.users
   Users_fully_twoSession_active [n] 
   # n.members = 5                             }
run TrafficTest4 for 1 but
   5 Name, 3 LinkIdent, 8 Link, 6 SessionIdent, 12 NetHdr,
   5 AcquireTable, 5 ForwardTable

pred TrafficTest5 [n: NetworkStateWithTraffic] {
   ! No_self_sessions [n]                      }
run TrafficTest5 for 1 but
   3 Name, 3 LinkIdent, 5 Link, 3 AcquireTable, 3 ForwardTable

pred TrafficTest6 [n: NetworkStateWithTraffic] {
   ! Sending_is_authentic [n]                  }
run TrafficTest6 for 1 but
   3 Name, 3 LinkIdent, 5 Link, 3 AcquireTable, 3 ForwardTable

pred TrafficTest7 [n: NetworkStateWithTraffic] {
   ! Receiving_is_authentic [n]                }
run TrafficTest7 for 1 but
   3 Name, 3 LinkIdent, 5 Link, 3 AcquireTable, 3 ForwardTable

pred TrafficTest8 [n: NetworkStateWithTraffic] {
   ! Session_identifiers_are_unique [n]        }
run TrafficTest8 for 1 but
   3 Name, 3 LinkIdent, 5 Link, 3 SessionIdent, 4 NetHdr,
   3 AcquireTable, 3 ForwardTable

pred TrafficTest9 [n: NetworkStateWithTraffic] {
   ! Traffic_is_symmetric [n]                  }
run TrafficTest9 for 1 but
   3 Name, 3 LinkIdent, 5 Link, 4 SessionIdent, 4 NetHdr,
   3 AcquireTable, 3 ForwardTable

pred TrafficTest10 [n: NetworkStateWithTraffic] {
   some n.effectivelyReachable
   No_self_sessions [n]                      
   Sending_is_authentic [n]                 
   Receiving_is_authentic [n]                 
   Session_identifiers_are_unique [n] 
   Traffic_is_symmetric [n]                     }
run TrafficTest10 for 1 but
   3 Name, 3 LinkIdent, 5 Link, 6 SessionIdent, 6 NetHdr,
   3 AcquireTable, 3 ForwardTable

/* --------------------------------------------------------------------
TESTS OF NETWORK BEHAVIOR
-------------------------------------------------------------------- */

pred BehaviorTest1 [n: NetworkStateWithTraffic] { 
   TrafficTest1 [n]
   Network_satisfies_communication_demands [n]  }
run BehaviorTest1 for 1 but
   4 Name, 2 LinkIdent, 6 Link, 12 SessionIdent, 12 NetHdr,
   4 AcquireTable, 4 ForwardTable

pred BehaviorTest2 [n: NetworkStateWithTraffic] { 
   TrafficTest2 [n]
   Only_authentic_traffic_delivered [n]         
   Network_satisfies_communication_demands [n]  }
run BehaviorTest2 for 1 but
   4 Name, 3 LinkIdent, 6 Link, 6 SessionIdent, 6 NetHdr,
   4 AcquireTable, 4 ForwardTable

pred BehaviorTest3 [n: NetworkStateWithTraffic] { 
   TrafficTest3 [n]
   Effective_reachability_is_symmetric [n]      
   Network_satisfies_communication_demands [n]  }
run BehaviorTest3 for 1 but
   4 Name, 2 LinkIdent, 4 Link, 6 SessionIdent, 12 NetHdr,
   4 AcquireTable, 4 ForwardTable

pred BehaviorTest4 [n: NetworkStateWithTraffic] { 
   !Network_satisfies_communication_demands [n] 
   ! Only_authentic_traffic_delivered [n]     
   ! Effective_reachability_is_symmetric [n]    }
run BehaviorTest4 for 1 but
   4 Name, 3 LinkIdent, 6 Link, 8 SessionIdent, 8 NetHdr,
   4 AcquireTable, 4 ForwardTable

pred BehaviorTest5 [n: NetworkStateWithTraffic] { 
   Network_satisfies_communication_demands [n] 
   ! Only_authentic_traffic_delivered [n]     
   ! Effective_reachability_is_symmetric [n]    }
run BehaviorTest5 for 1 but
   4 Name, 3 LinkIdent, 6 Link, 8 SessionIdent, 8 NetHdr,
   4 AcquireTable, 4 ForwardTable

pred BehaviorTest6 [n: NetworkStateWithTraffic] { 
      Effective_reachability_is_symmetric [n] 
   && ! Only_authentic_traffic_delivered [n]    }
run BehaviorTest6 for 1 but
   4 Name, 3 LinkIdent, 6 Link, 8 SessionIdent, 8 NetHdr,
   4 AcquireTable, 4 ForwardTable

/* --------------------------------------------------------------------
TESTS AND THEOREMS ABOUT EQUIVALENCE
-------------------------------------------------------------------- */

pred EquivalenceTest1 [disj n, n": NetworkState] {
   Networks_effectively_equivalent [n, n"] && n.links != n".links  }
run EquivalenceTest1 for 1 but 2 NetworkName, 2 NetworkState,
   4 Name, 3 LinkIdent, 8 Link, 4 AcquireTable, 4 ForwardTable

pred EquivalenceTest2 [disj n, n": NetworkState] {
   Networks_effectively_equivalent [n, n"] && n.members != n".members }
run EquivalenceTest2 for 1 but 2 NetworkName, 2 NetworkState,
   4 Name, 3 LinkIdent, 8 Link, 4 AcquireTable, 4 ForwardTable

assert Equivalence_reflexivity_theorem {
   all n: NetworkState | Networks_effectively_equivalent [n, n]  }
check Equivalence_reflexivity_theorem for 1 but
   6 Name, 4 LinkIdent, 8 Link, 2 SessionIdent, 2 NetHdr,
   6 AcquireTable, 6 ForwardTable

assert Equivalence_symmetry_theorem {
   all n, n": NetworkState |
      Networks_effectively_equivalent [n, n"] 
   => Networks_effectively_equivalent [n", n]  }
check Equivalence_symmetry_theorem for 1 but
   2 NetworkName, 2 NetworkState,
   6 Name, 4 LinkIdent, 8 Link, 2 SessionIdent, 2 NetHdr,
   6 AcquireTable, 6 ForwardTable

assert Equivalence_transitivity_theorem {
   all n0, n1, n2: NetworkState |
      Networks_effectively_equivalent [n0, n1] 
   && Networks_effectively_equivalent [n1, n2]
   => Networks_effectively_equivalent [n0, n2]  }
check Equivalence_transitivity_theorem for 1 but
   3 NetworkName, 3 NetworkState,
   6 Name, 4 LinkIdent, 8 Link, 2 SessionIdent, 2 NetHdr,
   6 AcquireTable, 6 ForwardTable
